import os
import pickle
import shutil
import numpy as np
import pandas as pd
import ncpi

# Path to the folder containing the processed data
sim_file_path = '/DATOS/pablomc/data/Hagen_model_v1'

# Path to the folder where the features will be saved
features_path = '/DATOS/pablomc/data'

# Set to True if features should be computed for the EEG data instead of the CDM data
compute_EEG = False

if __name__ == '__main__':
    # Create the FieldPotential object
    if compute_EEG:
        potential = ncpi.FieldPotential(kernel=False, nyhead=True, MEEG ='EEG')

    for method in ['catch22', 'power_spectrum_parameterization_1']:
        # Check if the features have already been computed
        folder = 'EEG' if compute_EEG else ''
        if os.path.isfile(os.path.join(features_path, method, folder, 'sim_X')):
            print(f'Features have already been computed for the method {method}.')
        else:
            print(f'Computing features for the method {method}.')
            # Create folders
            if not os.path.isdir(features_path):
                os.mkdir(features_path)
            if not os.path.isdir(os.path.join(features_path, method)):
                os.mkdir(os.path.join(features_path, method))
            if not os.path.isdir(os.path.join(features_path, method, 'tmp')):
                os.mkdir(os.path.join(features_path, method, 'tmp'))

            # Process files in the folder
            ldir = os.listdir(sim_file_path)
            for file in ldir:
                print(file)

                # CDM data
                if file[:3] == 'CDM':
                    CDM_data = pickle.load(open(os.path.join(sim_file_path,file),'rb'))

                    # Split CDM data into 10 chunks when computing EEGs to avoid memory issues
                    if compute_EEG:
                        # Check if the features have already been computed for this file
                        if os.path.isfile(os.path.join(features_path, method, 'tmp',
                                                       'sim_X_'+file.split('_')[-1]+'_0')) == False:
                            if len(CDM_data) > 10:
                                all_CDM_data = np.array_split(CDM_data, 10)
                            else:
                                all_CDM_data = [CDM_data]
                        else:
                            print(f'Features have already been computed for CDM data {file.split("_")[-1]}.')
                            continue
                    else:
                        all_CDM_data = [CDM_data]

                    all_features = []
                    for ii, data_chunk_1 in enumerate(all_CDM_data):
                        print(f'Computing features for CDM data chunk {ii+1}/{len(all_CDM_data)}')
                        # Computation of EEGs
                        if compute_EEG:
                            # Check if the features have already been computed for this chunk
                            if os.path.isfile(os.path.join(features_path, method, 'tmp',
                                                           'all_features_' + file.split('_')[-1] + '_' + str(ii))) == False:
                                print(f'Computing EEGs for CDM data chunk {ii+1}/{len(all_CDM_data)}')
                                all_data = np.zeros((len(data_chunk_1), 20, len(data_chunk_1[0])))
                                for k,CDM_data in enumerate(data_chunk_1):
                                    # print(f'EEG {k+1}/{len(data_chunk_1)}', end='\r')
                                    EEGs = potential.compute_MEEG(CDM_data)
                                    all_data[k,:,:] = EEGs
                            else:
                                print(f'Features have already been computed for CDM data chunk {ii+1}/{len(all_CDM_data)}')
                                continue
                        else:
                            all_data = [data_chunk_1]

                        # Get the features for each chunk
                        for jj, data_chunk_2 in enumerate(all_data):
                            # print(f'Chunk {jj+1}/{len(all_data)} of CDM_data {ii+1}/{len(all_CDM_data)}')
                            # Create a fake Pandas DataFrame (only Data and fs are relevant)
                            df = pd.DataFrame({'ID': np.zeros(len(data_chunk_2)),
                                               'Group': np.arange(len(data_chunk_2)),
                                               'Epoch': np.zeros(len(data_chunk_2)),
                                               'Sensor': np.zeros(len(data_chunk_2)),
                                               'Data': list(data_chunk_2)})
                            df.Recording = 'EEG' if compute_EEG else 'CDM'
                            df.fs = 1000. / 0.625 # samples/s

                            # Compute features
                            if method == 'catch22':
                                features = ncpi.Features(method='catch22')
                            elif method == 'power_spectrum_parameterization_1' or method == 'power_spectrum_parameterization_2':
                                # Parameters of the fooof algorithm
                                fooof_setup_sim = {'peak_threshold': 1.,
                                                   'min_peak_height': 0.,
                                                   'max_n_peaks': 5,
                                                   'peak_width_limits': (10., 50.)}
                                features = ncpi.Features(method='power_spectrum_parameterization',
                                                         params={'fs': df.fs,
                                                                 'fmin': 5.,
                                                                 'fmax': 200.,
                                                                 'fooof_setup': fooof_setup_sim,
                                                                 'r_squared_th':0.9})
                            elif method == 'fEI':
                                features = ncpi.Features(method='fEI',
                                                         params={'fs': df.fs,
                                                                 'fmin': 5.,
                                                                 'fmax': 150.,
                                                                 'fEI_folder': '../../../ncpi/Matlab'})

                            df = features.compute_features(df)

                            # Keep only the aperiodic exponent
                            if method == 'power_spectrum_parameterization_1':
                                df['Features'] = df['Features'].apply(lambda x: x[1])
                            # Keep aperiodic exponent, peak frequency, peak power, knee frequency, and mean power
                            if method == 'power_spectrum_parameterization_2':
                                df['Features'] = df['Features'].apply(lambda x: x[[1, 2, 3, 6, 11]])

                            # Append the feature dataframes to a list
                            all_features.append(df['Features'].tolist())

                        # Save the features to a tmp file
                        if compute_EEG:
                            pickle.dump(all_features, open(os.path.join(features_path, method, 'tmp',
                                                           'all_features_' + file.split('_')[-1] + '_' + str(ii)), 'wb'))
                            # Kill the process to clear memory and start again
                            if ii < len(all_CDM_data)-1:
                                os._exit(0)

                    if compute_EEG:
                        # Merge the features into a single list
                        all_features = []
                        for ii in range(len(all_CDM_data)):
                            feats = pickle.load(open(os.path.join(features_path, method, 'tmp',
                                                           'all_features_' + file.split('_')[-1] + '_' + str(ii)), 'rb'))
                            all_features.extend(feats)

                        # Remove feature files
                        for ii in range(len(all_CDM_data)):
                            os.remove(os.path.join(features_path, method, 'tmp',
                                                           'all_features_' + file.split('_')[-1] + '_' + str(ii)))

                        # Save the features to a file
                        print('\nSaving EEG features to files.')
                        for i in range(20):
                            elec_data = []
                            for j in range(len(all_features)):
                                elec_data.append(all_features[j][i])

                            pickle.dump(np.array(elec_data),open(os.path.join(features_path, method, 'tmp',
                                                          'sim_X_'+file.split('_')[-1]+'_'+str(i)), 'wb'))
                        # Kill the process
                        os._exit(0)

                    else:
                        df = all_features[0]

                        # Save the features to a file
                        pickle.dump(np.array(df),
                                    open(os.path.join(features_path, method, 'tmp',
                                                      'sim_X_'+file.split('_')[-1]), 'wb'))

                    # clear memory
                    del all_data, CDM_data, df, all_features

                # Theta data
                elif file[:5] == 'theta':
                    theta = pickle.load(open(os.path.join(sim_file_path, file), 'rb'))

                    # Save parameters to a file
                    pickle.dump(theta, open(os.path.join(features_path, method, 'tmp',
                                                         'sim_theta_'+file.split('_')[-1]), 'wb'))

            # Merge the features and parameters into single files
            print('\nMerging features and parameters into single files.')

            ldir = os.listdir(os.path.join(features_path, method, 'tmp'))
            theta_files = [file for file in ldir if file[:9] == 'sim_theta']
            num_files = len(theta_files)

            if compute_EEG:
                # Create EEG folder
                if not os.path.isdir(os.path.join(features_path, method, 'EEG')):
                    os.mkdir(os.path.join(features_path, method, 'EEG'))

                X = [[] for _ in range(20)]
                theta = []
                parameters = []

                for ii in range(int(num_files)):
                    data_theta = pickle.load(open(os.path.join(features_path, method, 'tmp','sim_theta_'+str(ii)), 'rb'))
                    theta.append(data_theta['data'])
                    parameters.append(data_theta['parameters'])

                    for jj in range(20):
                        data_X = pickle.load(open(os.path.join(features_path, method,
                                                               'tmp','sim_X_'+str(ii)+'_'+str(jj)), 'rb'))
                        X[jj].append(data_X)

                # Save features to files
                for jj in range(20):
                    pickle.dump(np.concatenate(X[jj]),
                                open(os.path.join(features_path, method, 'EEG', 'sim_X_'+str(jj)), 'wb'))

            else:
                X = []
                theta = []
                parameters = []

                for ii in range(int(num_files)):
                    data_theta = pickle.load(open(os.path.join(features_path, method, 'tmp','sim_theta_'+str(ii)), 'rb'))
                    theta.append(data_theta['data'])
                    parameters.append(data_theta['parameters'])
                    data_X = pickle.load(open(os.path.join(features_path, method, 'tmp', 'sim_X_' + str(ii)), 'rb'))
                    X.append(data_X)

                # Save features to files
                pickle.dump(np.concatenate(X), open(os.path.join(features_path, method, 'sim_X'), 'wb'))

            # Save the parameters to a file
            if os.path.isfile(os.path.join(features_path, method, 'sim_theta')) == False:
                th = {'data': np.concatenate(theta), 'parameters': parameters[0]}
                pickle.dump(th, open(os.path.join(features_path, method, 'sim_theta'), 'wb'))
                print(f"\nFeatures computed for {len(th['data'])} samples.")

            # Remove the 'tmp' folder
            shutil.rmtree(os.path.join(features_path, method, 'tmp'))


import os
import pickle
import numpy as np
import time
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from scipy.ndimage import gaussian_filter1d
from ncpi import tools

# Choose to either download data from Zenodo (True) or load it from a local path (False).
# Important: the zenodo downloads will take a while, so if you have already downloaded the data, set this to False and
# configure the zenodo_dir variable to point to the local path where the data is stored.
zenodo_dw_sim = True # simulation data

# Zenodo URL that contains the simulation data and ML models (used if zenodo_dw_sim is True)
zenodo_URL_sim = "https://zenodo.org/api/records/15351118"

# Paths to zenodo files
zenodo_dir_sim = "zenodo_sim_files"

# ML model used to compute the predictions (MLPRegressor or Ridge)
ML_model = 'MLPRegressor'

# Download simulation data and ML models
if zenodo_dw_sim:
    print('\n--- Downloading simulation data and ML models from Zenodo.')
    start_time = time.time()
    tools.download_zenodo_record(zenodo_URL_sim, download_dir=zenodo_dir_sim)
    end_time = time.time()
    print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")


def hellinger_distance(p, q):
    """
    Compute the Hellinger Distance between two discrete probability distributions P and Q.

    Parameters:
    p (np.array): Probability distribution P.
    q (np.array): Probability distribution Q.

    Returns:
    float: Hellinger Distance between P and Q.
    """
    # Ensure the inputs are numpy arrays
    p = np.asarray(p)
    q = np.asarray(q)

    # Normalize the distributions to ensure they sum to 1
    p = p / np.sum(p)
    q = q / np.sum(q)

    # Compute the Hellinger Distance
    return np.sqrt(np.sum((np.sqrt(p) - np.sqrt(q)) ** 2)) / np.sqrt(2)

# Set the random seed
np.random.seed(0)

# Load predictions and actual parameters
all_methods = ['catch22', 'catch22_psp_1', 'SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1',
               'SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1', 'MD_hrv_classic_pnn40',
               'power_spectrum_parameterization_1']

if ML_model == 'MLPRegressor':
    folder = 'MLP'
elif ML_model == 'Ridge':
    folder = 'Ridge'

all_preds = {}
all_theta = {}
for method in all_methods:
    try:
        with open(os.path.join(zenodo_dir_sim,'ML_models/held_out_data_models', folder, method,
                               'predictions'), 'rb') as file:
            all_preds[method] = np.array(pickle.load(file))
        with open(os.path.join(zenodo_dir_sim,'ML_models/held_out_data_models', 'datasets', method,
                               'held_out_dataset'), 'rb') as file:
            X_test, theta_test = pickle.load(file)
            all_theta[method] = np.array(theta_test)
    except:
        print('No data for method: ', method)

# Create a figure and set its properties
fig1 = plt.figure(figsize=(7.5, 6), dpi=300)
plt.rcParams.update({'font.size': 10, 'font.family': 'Arial'})
plt.rc('xtick', labelsize=8)
plt.rc('ytick', labelsize=8)

# Define a colormap
cmap = plt.colormaps['pink']

# Plot predictions vs actual parameters
all_errors = {'E_I': {}, 'tau_syn_exc': {}, 'tau_syn_inh': {}, 'J_syn_ext': {}}
for row in range(4):
    for col in range(6):
        ax = fig1.add_axes([0.12 + col * 0.15, 0.84 - row * 0.15, 0.09, 0.1])
        method = all_methods[col]

        try:
            # Compute E/I ratio
            if row == 0:
                pred_param = (all_preds[method][:,0]/all_preds[method][:,2]) /\
                             (all_preds[method][:,1]/all_preds[method][:,3])
                actual_param = (all_theta[method][:,0]/all_theta[method][:,2]) /\
                               (all_theta[method][:,1]/all_theta[method][:,3])
            else:
                pred_param = all_preds[method][:, row+3]
                actual_param = all_theta[method][:, row+3]

            # Remove nan values
            nan_pos = np.where(np.isnan(pred_param))[0]
            if len(nan_pos) > 0:
                pred_param = np.delete(pred_param, nan_pos)
                actual_param = np.delete(actual_param, nan_pos)

            # Define the range of the parameter
            if row == 0:
                min_ = 0.1
                max_ = 8.
            elif row == 1:
                min_ = 0.1
                max_ = 2.
            elif row == 2:
                min_ = 1.
                max_ = 8.
            elif row == 3:
                min_ = 10.
                max_ = 40.

            param_range = np.linspace(min_, max_, 8)

            # Use the same number of samples for each bin
            min_number = 10**10
            for ii,param in enumerate(param_range[:-1]):
                pos = np.where((actual_param >= param_range[ii]) & (actual_param < param_range[ii+1]))[0]
                if len(pos) < min_number:
                    min_number = len(pos)

            # Plots
            all_loc = []
            all_minmax = []
            abs_error = []
            for ii,param in enumerate(param_range[:-1]):
                pos = np.where((actual_param >= param_range[ii]) & (actual_param < param_range[ii+1]))[0]
                loc = (param_range[ii] + param_range[ii+1]) / 2
                all_loc.append(loc)

                # Choose the same number of samples for each bin
                if len(pos) > min_number:
                    pos = np.random.choice(pos, min_number, replace=False)

                # print(f'Method: {method}, row: {row}, col: {col}, param: {param}, number os samples: {len(pos)}')

                if len(pos) == 0:
                    continue

                # # Clip the data between the 15 % and 85 % quantiles
                if row == 0 and col == 3:
                   pp1 = 15
                   pp2 = 85
                # Clip the data between the 5 % and 95 % quantiles
                else:
                    pp1 = 5
                    pp2 = 95

                q1, q3 = np.percentile(pred_param[pos], [pp1, pp2])
                clipped_data = pred_param[pos][(pred_param[pos] >= q1) & (pred_param[pos] <= q3)]
                all_minmax.append([np.min(clipped_data), np.max(clipped_data)])

                # Violin plot
                violin = ax.violinplot(clipped_data, positions=[loc], widths=(max_ - min_)/9., showextrema=False)

                for pc in violin['bodies']:
                    pc.set_facecolor(cmap(ii / len(param_range)))
                    pc.set_edgecolor('black')
                    pc.set_alpha(0.8)
                    pc.set_linewidth(0.2)

                # Boxplot
                box = ax.boxplot(pred_param[pos], positions=[loc], showfliers=False,
                                 widths=(max_ - min_)/9., patch_artist=True,
                                 medianprops=dict(color='red', linewidth=0.8),
                                 whiskerprops=dict(color='black', linewidth=0.5),
                                 capprops=dict(color='black', linewidth=0.5),
                                 boxprops=dict(linewidth=0.5, facecolor=(0, 0, 0, 0)))

                for patch in box['boxes']:
                    patch.set_linewidth(0.2)

                # Compute absolute errors
                error = np.abs(pred_param[pos] - actual_param[pos])
                abs_error.extend(error)

            # Store the errors
            all_errors[list(all_errors.keys())[row]][method] = abs_error

            # ticks
            ax.set_xticks(all_loc[::2 if row < 3 else 3])
            ax.set_xticklabels(['%.1f' % x for x in all_loc[::2 if row < 3 else 3]])
            all_minmax = np.array(all_minmax)
            ax.set_yticks(np.linspace(np.min(all_minmax), np.max(all_minmax), 3))
            ax.set_yticklabels(['%.1f' % x for x in np.linspace(np.min(all_minmax), np.max(all_minmax), 3)])

            # x-/y-labels
            if row == 3:
                ax.set_xlabel('actual', fontsize = 8)
            if col == 0:
                ax.set_ylabel('predicted', fontsize = 8)
                ax.yaxis.set_label_coords(-0.5, 0.5)

            # titles
            if row == 0:
                if col == 0:
                    ax.set_title(r'$catch22$', fontsize = 8)
                elif col == 1:
                    ax.set_title(r'$catch22 + $'+  r'$1/f$'+' '+r'$slope$', fontsize = 8)
                elif col == 2:
                    ax.set_title(r'$dfa$', fontsize = 8)
                elif col == 3:
                    ax.set_title(r'$rs\ range$', fontsize = 8)
                elif col == 4:
                    ax.set_title(r'$high\ fluct.$', fontsize = 8)
                elif col == 5:
                    ax.set_title(r'$1/f$'+' '+r'$slope$', fontsize = 8)

            # limits
            ax.set_xlim([min_, max_])

        except:
            continue

# Show parameter labels
for row in range(4):
    ax = fig1.add_axes([0.01, 0.84 - row * 0.15, 0.04, 0.1])
    ax.axis('off')

    if row == 0:
        ax.text(0.5, 0.5, r'$E/I$', fontsize=8, ha='center')
    elif row == 1:
        ax.text(0.5, 0.5, r'$\tau_{syn}^{exc}$', fontsize=8, ha='center')
    elif row == 2:
        ax.text(0.5, 0.5, r'$\tau_{syn}^{inh}$', fontsize=8, ha='center')
    else:
        ax.text(0.5, 0.5, r'$J_{syn}^{ext}$', fontsize=8, ha='center')

# Plot errors
colors = ['#FFC0CB', '#FF69B4', '#00FF00', '#32CD32', '#228B22', '#006400']
cmap = LinearSegmentedColormap.from_list('custom_cmap', colors, N=6)

for col in range(4):
    ax = fig1.add_axes([0.08 + col * 0.24, 0.07, 0.18, 0.18])
    all_hist = []

    # Define bins to compute histograms
    if col == 0:
        bins = np.linspace(0, 10, 15)
    elif col == 1:
        bins = np.linspace(0, 1.5, 15)
    elif col == 2:
        bins = np.linspace(0, 5, 15)
    else:
        bins = np.linspace(0, 40, 15)

    for ii,method in enumerate(all_methods):
        try:
            # titles
            if ii == 0:
                label = r'$catch22$'
            elif ii == 1:
                label = r'$ch22 + $'+' '+r'slp'
            elif ii == 2:
                label = r'$dfa$'
            elif ii == 3:
                label = r'$rs\ range$'
            elif ii == 4:
                label = r'$high\ fluct.$'
            elif ii == 5:
                label = r'$1/f$'+' '+r'$slope$'

            # Compute histogram
            hist, bin_edges = np.histogram(all_errors[list(all_errors.keys())[col]][method], bins=bins, density=True)

            # Smooth the histogram using a Gaussian filter
            smoothed_hist = gaussian_filter1d(hist, sigma=1)
            all_hist.append(smoothed_hist)
            ax.plot(bin_edges[:-1], smoothed_hist, label=label, color=colors[ii], alpha=0.4, linewidth = 1.5)

        except:
            continue

    # Compute pairwise Hellinger distances
    HD_ = []
    for x in range(6):
        for y in range(x+1, 6):
            HD_.append(hellinger_distance(all_hist[x], all_hist[y]))
            # print('Hellinger distance between', all_methods[x], 'and', all_methods[y], ':', HD_[-1])

    # Show distances
    # Between catch22 and catch22 + slope
    ax.text(0.2, 0.9 if col == 0 else 0.25, r'$D_{H,1}=$'+' %.2f' % HD_[0], fontsize=6, ha='center',
            transform=ax.transAxes, color = colors[1])
    # Between all the remaining methods
    ax.text(0.2, 0.15, r'$D_{H,2}=$'+' %.2f' % np.mean(HD_[9:]), fontsize=6, ha='center', transform=ax.transAxes,
            color = colors[5])
    # Between catch22 and all the remaining methods
    ax.text(0.2, 0.05, r'$D_{H,3}=$'+' %.2f' % np.mean(HD_[1:5]), fontsize=6, ha='center', transform=ax.transAxes,
            color = 'k')

    # legend
    if col == 0:
        ax.legend(loc='upper right', fontsize=6, handletextpad=0.2, borderpad=0.2, labelspacing=0.2)

    # labels
    if col == 0:
        ax.set_ylabel('probability density', fontsize=8)
    ax.set_xlabel('absolute error', fontsize=8)

    # titles
    if col == 0:
        ax.set_title(r'$E/I$', fontsize=10)
    elif col == 1:
        ax.set_title(r'$\tau_{syn}^{exc}$', fontsize=10)
    elif col == 2:
        ax.set_title(r'$\tau_{syn}^{inh}$', fontsize=10)
    else:
        ax.set_title(r'$J_{syn}^{ext}$', fontsize=10)

    # limits
    if col == 0:
        ax.set_xlim([0, 12])

# Plot letters
ax = fig1.add_axes([0., 0., 1., 1.])
ax.axis('off')
ax.text(0.005, 0.97, 'A', fontsize=12, fontweight='bold')
ax.text(0.01, 0.28, 'B', fontsize=12, fontweight='bold')

# Save the figure
plt.savefig(f'sim_predictions_{ML_model}.png', bbox_inches='tight')
# plt.show()


import os
import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
from ncpi import tools


class FieldPotential:
    def __init__(self, kernel=True, nyhead=False, MEEG=None, CDM_shape=15500):
        """
        Initialize the FieldPotential object.

        Parameters
        ----------
        kernel: bool
            Choose to compute kernels.

        nyhead: bool
            Use the NYHeadModel for computing EEG.

        MEEG: str
            Prepare data structures for computing EEG or MEG signals. Options are:
                'EEG': Compute EEG.
                'MEG': Compute MEG.

        CDM_shape: int
            Number of time points in the CDM.
        """

        if kernel:
            # Check that the required modules are installed
            if not tools.ensure_module("lfpykernels"):
                raise ImportError("lfpykernels is required for computing kernels but is not installed.")
            self.KernelApprox = tools.dynamic_import("lfpykernels",
                                                     "KernelApprox")
            self.GaussCylinderPotential = tools.dynamic_import("lfpykernels",
                                                               "GaussCylinderPotential")
            self.KernelApproxCurrentDipoleMoment = tools.dynamic_import("lfpykernels",
                                                                        "KernelApproxCurrentDipoleMoment")

            # Is LFPy import really needed?
            if not tools.ensure_module("LFPy"):
                raise ImportError("LFPy is required for computing kernels but is not installed.")

            if not tools.ensure_module("neuron"):
                raise ImportError("neuron is required for computing kernels but is not installed.")
            self.neuron = tools.dynamic_import("neuron")

            if not tools.ensure_module("h5py"):
                raise ImportError("h5py is required for computing kernels but is not installed.")
            self.h5py = tools.dynamic_import("h5py")

            # Initialize dictionary for storing kernels
            self.H_YX = dict()

        if nyhead:
            # Check that the required module is installed
            if not tools.ensure_module("lfpykit"):
                raise ImportError("lfpykit is required for computing the NYHeadModel but is not installed.")
            self.NYHeadModel = tools.dynamic_import("lfpykit.eegmegcalc",
                                                    "NYHeadModel")

            # Initialize the head model
            self.nyhead = self.NYHeadModel()

        # Import InfiniteHomogeneousVolCondMEG
        if MEEG is not None:
            self.IHVCMEG = tools.dynamic_import("LFPy", "InfiniteHomogeneousVolCondMEG")

        # Allocate memory for the transformation matrix and MEEG
        if MEEG == 'EEG':
            self.M = np.zeros((231,3))
            self.MEEG = np.zeros((231,CDM_shape))
        elif MEEG == 'MEG':
            self.M = np.zeros((1,3,3))
            self.MEEG = np.zeros((1,3,CDM_shape))


    def create_kernel(self, MC_folder, output_path, params, biophys, dt, tstop, electrodeParameters=None, CDM=True):
        """
        Create kernels from multicompartment neuron network descriptions.

        Parameters
        ----------
        MC_folder: str
            Path to the folder containing the multicompartment neuron network descriptions.
        output_path: str
            Path to the output folder.
        params: module
            `<module 'example_network_parameters'>`
        biophys: list
            List of biophysical membrane properties.
        dt: float
            Time step.
        tstop: float
            Simulation time.
        electrodeParameters: dict
            Electrode parameters. If None, no LFP is computed.
        CDM: bool
            Compute the current dipole moment.

        Returns
        -------
        H_YX: dict
            Dictionary containing the kernels.
        """

        # Check that folder exists
        if not os.path.exists(MC_folder):
            raise FileNotFoundError(f"{MC_folder} not found.")

        # Check that the output path exists
        if not os.path.exists(output_path):
            raise FileNotFoundError(f"{output_path} not found.")

        # Add some checks for the parameters?

        # Check that biophys is a list
        if not isinstance(biophys, list):
            raise TypeError("biophys must be a list.")

        # Check that dt is > 0
        if dt <= 0:
            raise ValueError("dt must be > 0.")

        # Check that tstop is > 0
        if tstop <= 0:
            raise ValueError("tstop must be > 0.")

        # Check that electrodeParameters is a dictionary
        if electrodeParameters is not None and not isinstance(electrodeParameters, dict):
            raise TypeError("electrodeParameters must be a dictionary.")

        # Check that CDM is a boolean
        if not isinstance(CDM, bool):
            raise TypeError("CDM must be a boolean.")

        # Update paths of cellParameters and morphologies
        params.cellParameters['templatefile'] = os.path.join(MC_folder, params.cellParameters['templatefile'])
        params.morphologies = [os.path.join(MC_folder, m) for m in params.morphologies]

        # Recompile mod files if needed
        mech_loaded = self.neuron.load_mechanisms(os.path.join(MC_folder, 'mod'))
        if not mech_loaded:
            os.system(f'cd {os.path.join(MC_folder, "mod")} && nrnivmodl && cd -')
            self.neuron.load_mechanisms(os.path.join(MC_folder, "mod"))

        # Presynaptic activation time
        t_X = params.transient

        # Synapse max. conductance (function, mean, st.dev., min.):
        weights = [[params.MC_params['weight_EE'], params.MC_params['weight_IE']],
                   [params.MC_params['weight_EI'], params.MC_params['weight_II']]]

        # # Define biophysical membrane properties
        set_biophys = [globals()[func_name] for func_name in biophys]

        # Class RecExtElectrode/PointSourcePotential parameters:
        if electrodeParameters is not None:
            for key in ['r', 'n', 'N', 'method']:
                del electrodeParameters[key]

        # Predictor assuming planar disk source elements convolved with Gaussian
        # along z-axis
        probes = []
        if electrodeParameters is not None:
            gauss_cyl_potential = self.GaussCylinderPotential(
                cell=None,
                z=electrodeParameters['z'],
                sigma=electrodeParameters['sigma'],
                R=params.populationParameters['pop_args']['radius'],
                sigma_z=params.populationParameters['pop_args']['scale'],
            )
            probes.append(gauss_cyl_potential)


        # Set up recording of current dipole moments.
        if CDM:
            current_dipole_moment = self.KernelApproxCurrentDipoleMoment(cell=None)
            probes.append(current_dipole_moment)

        # Compute average firing rate of presynaptic populations X
        mean_nu_X = compute_mean_nu_X(params, output_path, params.transient)

        # Compute kernels
        for i, (X, N_X) in enumerate(zip(params.population_names,
                                         params.population_sizes)):
            for j, (Y, N_Y, morphology) in enumerate(zip(params.population_names,
                                                         params.population_sizes,
                                                         params.morphologies)):
                # Extract median soma voltages from actual network simulation and
                # assume this value corresponds to Vrest.
                with self.h5py.File(os.path.join(output_path, 'somav.h5'
                                            ), 'r') as f:
                    Vrest = np.median(f[Y][()][:, 200:])

                cellParameters = deepcopy(params.cellParameters)
                cellParameters.update(dict(
                    morphology=morphology,
                    custom_fun=set_biophys,
                    custom_fun_args=[dict(Vrest=Vrest), dict(Vrest=Vrest)],
                ))

                # Some inputs must be lists
                synapseParameters = [
                    dict(weight=weights[ii][j],
                         syntype='Exp2Syn',
                         **params.synapseParameters[ii][j])
                    for ii in range(len(params.population_names))]
                synapsePositionArguments = [
                    params.synapsePositionArguments[ii][j]
                    for ii in range(len(params.population_names))]

                # Create kernel approximator object
                kernel = self.KernelApprox(
                    X=params.population_names,
                    Y=Y,
                    N_X=np.array(params.population_sizes),
                    N_Y=N_Y,
                    C_YX=np.array(params.connectionProbability[i]),
                    cellParameters=cellParameters,
                    populationParameters=params.populationParameters['pop_args'],
                    multapseFunction=params.multapseFunction,
                    multapseParameters=[params.multapseArguments[ii][j] for ii in range(len(params.population_names))],
                    delayFunction=params.delayFunction,
                    delayParameters=[params.delayArguments[ii][j] for ii in range(len(params.population_names))],
                    synapseParameters=synapseParameters,
                    synapsePositionArguments=synapsePositionArguments,
                    extSynapseParameters=params.extSynapseParameters,
                    nu_ext=1000. / params.netstim_interval,
                    n_ext=params.MC_params['n_ext'][j],
                    nu_X=mean_nu_X,
                )

                # make kernel predictions
                self.H_YX['{}:{}'.format(Y, X)] = kernel.get_kernel(
                    probes=probes,
                    Vrest=Vrest, dt=dt, X=X, t_X=t_X, tau=params.tau,
                    g_eff=params.MC_params['g_eff'],
                )

        return self.H_YX


    def compute_MEEG(self, CDM, location=None, cMEG = False):
        """
        Compute EEG/MEG from a current dipole moment using the NYHeadModel (EEG) or a homogeneous, isotropic, linear
        volume conductor model (MEG).

        Warning: The MEG calculation has not yet been tested and is currently limited to use with a single
        sensor location.

        Parameters
        ----------
        CDM: np.array
            Current dipole moment
        location: np.array
            Location of the dipole moment
        cMEG: bool
            Compute MEG instead of EEG.

        Returns
        -------
        all_MEEG: list of np.array
            MEEG at the electrode/sensor locations
        """

        debug = False

        # Reformat the CDM to be a 3DxN array, where N is the number of time points
        p = np.zeros((3, len(CDM)))
        p[2, :] = CDM

        # If location is provided, compute MEEG at that location
        if location is not None:
            if cMEG == False:
                # Set the dipole location
                self.nyhead.set_dipole_pos(location)

                # Get the transformation matrix
                self.M = self.nyhead.get_transformation_matrix()

                # Rotate current dipole moment to be oriented along the normal vector of cortex
                p = self.nyhead.rotate_dipole_to_surface_normal(p)

                # Compute EEG
                self.MEEG = self.M @ p  # (mV)

                # Get the closest electrode idx to dipole location
                dist, closest_elec_idx = self.nyhead.find_closest_electrode()
                all_MEEG = self.MEEG[closest_elec_idx, :]

            # Warning: this code snippet has not been tested yet!
            else:
                # Instantiate MEG object
                meg = self.IHVCMEG(location)

                # Get the transformation matrix
                self.M = meg.get_transformation_matrix(location)

                # Compute the magnetic signal in a single sensor location
                self.MEEG = self.M @ p
                all_MEEG = self.MEEG[0]

        # If location is not provided, place the dipole at the different locations of the 10-20 EEG setup. We assume
        # that each dipole is working independently, i.e., the EEG at each electrode is computed separately.
        elif location is None and cMEG == False:
            all_MEEG = np.zeros((20,self.MEEG.shape[1]))
            locations = [np.array([-25, 65, 0]), # Fp1
                        np.array([25, 65, 0]),  # Fp2
                        np.array([-50, 36, -10]), # F7
                        np.array([-39, 36, 36]),  # F3
                        np.array([-3, 36, 56]),  # Fz
                        np.array([39, 36, 36]),  # F4
                        np.array([50, 36, -10]),  # F8
                        np.array([-68, -20, 0]),  # T3
                        np.array([-46, -15, 48]),  # C3
                        np.array([-3, -19, 76]),  # Cz
                        np.array([46, -15, 48]),  # C4
                        np.array([68, -20, 0]),  # T4
                        np.array([-57, -61, -17]),  # T5
                        np.array([-40, -55, 50]),  # P3
                        np.array([-7, -52, 72]),  # Pz
                        np.array([40, -55, 50]),  # P4
                        np.array([57, -61, -17]),  # T6
                        np.array([-32, -90, 18]),  # O1
                        np.array([-3, -92, 20]),  # Oz
                        np.array([32, -90, 18])]  # O2

            labels = ['Fp1', 'Fp2', 'F7', 'F3', 'Fz', 'F4', 'F8', 'T3', 'C3', 'Cz',
                      'C4', 'T4', 'T5', 'P3', 'Pz', 'P4', 'T6', 'O1', 'Oz', 'O2']

            # Plot electrode positions
            if debug:
                x_lim = [-100, 100]
                y_lim = [-130, 100]
                z_lim = [-160, 120]

                fig = plt.figure(figsize=(7.5, 3.), dpi=300)
                ax1 = fig.add_axes([0.15, 0.15, 0.2, 0.7],
                                   xlabel="x (mm)", ylabel='y (mm)', xlim=x_lim, ylim=y_lim)
                ax2 = fig.add_axes([0.45, 0.15, 0.2, 0.7],
                                   xlabel="x (mm)", ylabel='z (mm)', xlim=x_lim, ylim=z_lim)
                ax3 = fig.add_axes([0.75, 0.15, 0.2, 0.7],
                                   xlabel="y (mm)", ylabel='z (mm)', xlim=y_lim, ylim=z_lim)

                for i, location in enumerate(locations):
                    # Set the dipole location
                    self.nyhead.set_dipole_pos(location)

                    # Plot the head model and the dipole location
                    if i == 0:
                        threshold = 2
                        xz_plane_idxs = np.where(np.abs(self.nyhead.cortex[1, :] - 0) < threshold)[0]
                        xy_plane_idxs = np.where(np.abs(self.nyhead.cortex[2, :] - 0) < threshold)[0]
                        yz_plane_idxs = np.where(np.abs(self.nyhead.cortex[0, :] - 25) < threshold)[0]

                        ax1.scatter(self.nyhead.cortex[0, xy_plane_idxs], self.nyhead.cortex[1, xy_plane_idxs], s=0.05, color = 'r')
                        ax2.scatter(self.nyhead.cortex[0, xz_plane_idxs], self.nyhead.cortex[2, xz_plane_idxs], s=0.05, color = 'r')
                        ax3.scatter(self.nyhead.cortex[1, yz_plane_idxs], self.nyhead.cortex[2, yz_plane_idxs], s=0.05, color = 'r')

                    ax1.plot(self.nyhead.dipole_pos[0], self.nyhead.dipole_pos[1], 'o', ms=6, color='orange', zorder=1000)
                    ax2.plot(self.nyhead.dipole_pos[0], self.nyhead.dipole_pos[2], 'o', ms=6, color='orange', zorder=1000)
                    ax3.plot(self.nyhead.dipole_pos[1], self.nyhead.dipole_pos[2], 'o', ms=6, color='orange', zorder=1000)

                # Add labels
                for i, txt in enumerate(labels):
                    ax1.annotate(txt, (locations[i][0]+6, locations[i][1]), fontsize=6, color='k')
                    ax2.annotate(txt, (locations[i][0]+6, locations[i][2]), fontsize=6, color='k')
                    ax3.annotate(txt, (locations[i][1]+6, locations[i][2]), fontsize=6, color='k')

                plt.show()

            for ii,location in enumerate(locations):
                # Set the dipole location
                self.nyhead.set_dipole_pos(location)

                # Get the transformation matrix
                self.M = self.nyhead.get_transformation_matrix()

                # Rotate current dipole moment to be oriented along the normal vector of cortex
                p = self.nyhead.rotate_dipole_to_surface_normal(p)

                # Compute EEG
                self.MEEG = self.M @ p  # (mV)

                # Get the closest electrode idx to dipole location
                dist, closest_elec_idx = self.nyhead.find_closest_electrode()
                all_MEEG[ii,:] = self.MEEG[closest_elec_idx, :]

            # Delete variables
            # del p, dist, closest_elec_idx

        else:
            raise ValueError("Configuration (location/cMEG) not recognized.")


        return all_MEEG


    def compute_proxy(self, method, sim_data, sim_step):
        """
        Compute a proxy for the extracellular signal by combining variables directly measured from network simulations.
        Warning: not tested yet!

        Parameters
        ----------
        method: str
            Method to compute the proxy. Options are:
            - 'FR': firing rate averaged over all neurons in the population.
            - 'AMPA': sum of AMPA currents.
            - 'GABA': sum of GABA currents.
            - 'Vm': sum of membrane potentials.
            - 'I': sum of synaptic currents.
            - 'I_abs': sum of their absolute values.
            - 'LRWS': LFP reference weighted sum.
            - 'ERWS1': EEG reference weighted sum 1 (non-causal)
            - 'ERWS2': EEG reference weighted sum 2 (non-causal)

        sim_data: dict
            Dictionary containing the simulation data. The keys are 'FR', 'AMPA', 'GABA', 'Vm', 'nu_ext'. Each key
            contains a 2D array with the data for each neuron in the population.

        sim_step: float
            Simulation time step in ms.

        Returns
        -------
        proxy: np.array
            Proxy for the extracellular signal.
        """

        if method == 'FR':
            proxy = np.mean(sim_data['FR'], axis = 0)
        elif method == 'AMPA':
            proxy = np.sum(sim_data['AMPA'], axis = 0)
        elif method == 'GABA':
            proxy = -np.sum(sim_data['GABA'], axis = 0)
        elif method == 'Vm':
            proxy = np.sum(sim_data['Vm'], axis = 0)
        elif method == 'I':
            proxy = np.sum(sim_data['AMPA'] + sim_data['GABA'], axis = 0)
        elif method == 'I_abs':
            proxy = np.sum(np.abs(sim_data['AMPA']) + np.abs(sim_data['GABA']), axis = 0)
        elif method == 'LRWS':
            delay = int(6.0 / sim_step)
            AMPA_delayed = np.array([roll_with_zeros(sim_data['AMPA'][i],delay) for i in range(sim_data['AMPA'].shape[0])])
            proxy = np.sum(AMPA_delayed - 1.65 * sim_data['GABA'], axis=0)
        elif method == 'ERWS1':
            AMPA_delay = -int(0.9 / sim_step)
            GABA_delay = int(2.3 / sim_step)
            AMPA_delayed = np.array([roll_with_zeros(sim_data['AMPA'][i],
                                                     AMPA_delay) for i in range(sim_data['AMPA'].shape[0])])
            GABA_delayed = np.array([roll_with_zeros(sim_data['GABA'][i],
                                                     GABA_delay) for i in range(sim_data['GABA'].shape[0])])
            proxy = np.sum(AMPA_delayed - 0.3 * GABA_delayed, axis=0)
        elif method == 'ERWS2':
            coeff = [-0.6,  0.1, -0.4, -1.9,  0.6,  3.0 ,  1.4,  1.7,  0.2]
            AMPA_delay = int(coeff[0] * np.power(sim_data['nu_ext'], -coeff[1]) + coeff[2])
            GABA_delay = int(coeff[3] * np.power(sim_data['nu_ext'], -coeff[4]) + coeff[5])
            alpha = coeff[6] * np.power(sim_data['nu_ext'], -coeff[7]) + coeff[8]
            AMPA_delayed = np.array([roll_with_zeros(sim_data['AMPA'][i],
                                                     AMPA_delay) for i in range(sim_data['AMPA'].shape[0])])
            GABA_delayed = np.array([roll_with_zeros(sim_data['GABA'][i],
                                                     GABA_delay) for i in range(sim_data['GABA'].shape[0])])
            proxy = np.sum(AMPA_delayed - alpha * GABA_delayed, axis=0)
        else:
            raise ValueError(f"Method {method} not recognized.")

        return proxy


def roll_with_zeros(arr, shift):
    """
    Roll an array with zeros.
    Parameters
    ----------
    arr: np.array
        Array to be rolled.
    shift: int
        Number of positions to shift the array.
    Returns
    -------
    result: np.array
        Shifted array.

    """
    result = np.zeros_like(arr)
    if shift > 0:
        result[shift:] = arr[:-shift]
    elif shift < 0:
        result[:shift] = arr[-shift:]
    else:
        result[:] = arr
    return result


"""
The following methods for setting up the neuron model were downloaded from the LFPykernels repository: 
https://github.com/LFPy/LFPykernels/blob/main/examples/example_network_methods.py
Copyright (C) 2021 https://github.com/espenhgn
"""

def set_active( cell, Vrest):
    """Insert HH and Ih channels across cell sections

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('hh')
        sec.insert('Ih')
        if sec.name().rfind('soma') >= 0:
            sec.gnabar_hh = 0.12
            sec.gkbar_hh = 0.036
            sec.gl_hh = 0.0003
            sec.el_hh = -54.3
            sec.ena = 50
            sec.ek = -77

            sec.gIhbar_Ih = 0.002

        if sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            # set HH channel conductancesto 10% of default in apical dendrite
            sec.gnabar_hh = 0.012
            sec.gkbar_hh = 0.0036
            sec.gl_hh = 0.0003
            sec.el_hh = -54.3
            sec.ena = 50
            sec.ek = -77

            # set higher Ih-conductance in apical dendrite
            sec.gIhbar_Ih = 0.01


def set_passive(cell, Vrest):
    """Insert passive leak channel across sections

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.g_pas = 0.0003
        sec.e_pas = Vrest


def set_Ih( cell, Vrest):
    """Insert passive leak and voltage-gated Ih across sections

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.insert('Ih')
        sec.e_pas = Vrest
        sec.g_pas = 0.0003
        if sec.name().rfind('soma') >= 0:
            sec.gIhbar_Ih = 0.002
        elif sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.gIhbar_Ih = 0.01


def set_Ih_linearized( cell, Vrest):
    """Insert passive leak and linearized Ih.

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.insert('Ih_linearized_v2')
        sec.e_pas = Vrest
        sec.g_pas = 0.0003
        sec.V_R_Ih_linearized_v2 = Vrest
        if sec.name().rfind('soma') >= 0:
            sec.gIhbar_Ih_linearized_v2 = 0.002
        elif sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.gIhbar_Ih_linearized_v2 = 0.01


def set_pas_hay2011( cell, Vrest):
    """Insert passive leak as in Hay2011 model

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('pas')
        if sec.name().rfind('soma') >= 0:
            sec.g_pas = 0.0000338
            sec.e_pas = -90

        if sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.g_pas = 0.0000589
            sec.e_pas = -90


def set_active_hay2011( cell, Vrest):
    """Insert passive leak, Ih, NaTa_t and SKv3_1 channels as in Hay 2011 model

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential (not used)
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.insert('Ih')
        sec.insert('NaTa_t')
        sec.insert('SKv3_1')
        sec.ena = 50
        sec.ek = -85
        if sec.name().rfind('soma') >= 0:
            sec.gNaTa_tbar_NaTa_t = 2.04
            sec.gSKv3_1bar_SKv3_1 = 0.693
            sec.g_pas = 0.0000338
            sec.e_pas = -90
            sec.gIhbar_Ih = 0.0002

        if sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.gNaTa_tbar_NaTa_t = 0.0213
            sec.gSKv3_1bar_SKv3_1 = 0.000261
            sec.g_pas = 0.0000589
            sec.e_pas = -90
            sec.gIhbar_Ih = 0.0002 * 10


def set_frozen_hay2011( cell, Vrest):
    """Set passive leak and linear passive-frozen versions of Ih, NaTa_t and
    SKv3_1 channels from Hay 2011 model

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.insert('NaTa_t_frozen')
        sec.insert('SKv3_1_frozen')
        sec.insert('Ih_linearized_v2_frozen')
        sec.e_pas = Vrest
        sec.V_R_NaTa_t_frozen = Vrest
        sec.V_R_SKv3_1_frozen = Vrest
        sec.V_R_Ih_linearized_v2_frozen = Vrest
        sec.ena = Vrest  # 50
        sec.ek = Vrest  # -85
        if sec.name().rfind('soma') >= 0:
            sec.gNaTa_tbar_NaTa_t_frozen = 2.04
            sec.gSKv3_1bar_SKv3_1_frozen = 0.693
            sec.g_pas = 0.0000338
            sec.gIhbar_Ih_linearized_v2_frozen = 0.0002
        elif sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.gNaTa_tbar_NaTa_t_frozen = 0.0213
            sec.gSKv3_1bar_SKv3_1_frozen = 0.000261
            sec.g_pas = 0.0000589
            sec.gIhbar_Ih_linearized_v2_frozen = 0.0002 * 10


def set_frozen_hay2011_no_Ih( cell, Vrest):
    """
    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.insert('NaTa_t_frozen')
        sec.insert('SKv3_1_frozen')
        sec.e_pas = Vrest
        sec.V_R_NaTa_t_frozen = Vrest
        sec.V_R_SKv3_1_frozen = Vrest
        sec.ena = Vrest  # 50
        sec.ek = Vrest  # -85
        if sec.name().rfind('soma') >= 0:
            sec.gNaTa_tbar_NaTa_t_frozen = 2.04
            sec.gSKv3_1bar_SKv3_1_frozen = 0.693
            sec.g_pas = 0.0000338
        elif sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.gNaTa_tbar_NaTa_t_frozen = 0.0213
            sec.gSKv3_1bar_SKv3_1_frozen = 0.000261
            sec.g_pas = 0.0000589


def set_Ih_hay2011( cell, Vrest):
    """
    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.insert('NaTa_t_frozen')
        sec.insert('SKv3_1_frozen')
        sec.insert('Ih')
        sec.e_pas = Vrest
        sec.ena = Vrest  # 50
        sec.V_R_NaTa_t_frozen = Vrest
        sec.V_R_SKv3_1_frozen = Vrest
        sec.ek = Vrest  # -85
        if sec.name().rfind('soma') >= 0:
            sec.gNaTa_tbar_NaTa_t_frozen = 2.04
            sec.gSKv3_1bar_SKv3_1_frozen = 0.693
            sec.g_pas = 0.0000338
            sec.gIhbar_Ih = 0.0002
        elif sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.gNaTa_tbar_NaTa_t_frozen = 0.0213
            sec.gSKv3_1bar_SKv3_1_frozen = 0.000261
            sec.g_pas = 0.0000589
            sec.gIhbar_Ih = 0.0002 * 10


def set_Ih_linearized_hay2011( cell, Vrest):
    """
    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.insert('NaTa_t_frozen')
        sec.insert('SKv3_1_frozen')
        sec.insert('Ih_linearized_v2')
        sec.e_pas = Vrest
        sec.V_R_Ih_linearized_v2 = Vrest
        sec.V_R_NaTa_t_frozen = Vrest
        sec.V_R_SKv3_1_frozen = Vrest
        sec.ena = Vrest  # 50
        sec.ek = Vrest  # -85

        if sec.name().rfind('soma') >= 0:
            sec.gNaTa_tbar_NaTa_t_frozen = 2.04
            sec.gSKv3_1bar_SKv3_1_frozen = 0.693
            sec.g_pas = 0.0000338
            sec.gIhbar_Ih_linearized_v2 = 0.0002
        elif sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.gNaTa_tbar_NaTa_t_frozen = 0.0213
            sec.gSKv3_1bar_SKv3_1_frozen = 0.000261
            sec.g_pas = 0.0000589
            sec.gIhbar_Ih_linearized_v2 = 0.0002 * 10


def set_V_R_Ih_linearized_v2( cell, Vrest):
    """
    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """

    import neuron

    for sec in cell.template.all:
        if neuron.h.ismembrane("Ih_linearized_v2", sec=sec):
            sec.V_R_Ih_linearized_v2 = Vrest
        elif neuron.h.ismembrane("Ih_linearized_v2_frozen", sec=sec):
            sec.V_R_Ih_linearized_v2_frozen = Vrest


def set_V_R( cell, Vrest):
    """
    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """

    import neuron

    ion_channels = [
        "Ih_linearized_v2",
        "Ih_linearized_v2_frozen",
        "Ca_LVAst_frozen",
        "Ca_HVA_frozen",
        "SKv3_1_frozen",
        "K_Tst_frozen",
        "K_Pst_frozen",
        "Nap_Et2_frozen",
        "Nap_Et2_linearized",
        "NaTa_t_frozen",
        "Im_frozen"
    ]
    for sec in cell.template.all:
        for ion in ion_channels:
            if neuron.h.ismembrane(ion, sec=sec):
                setattr(sec, f'V_R_{ion}', Vrest)


def make_cell_uniform( cell, Vrest=-65):
    """
    Adjusts e_pas to enforce a uniform resting membrane potential at Vrest

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential
    """

    import neuron

    neuron.h.t = 0
    neuron.h.finitialize(Vrest)
    neuron.h.fcurrent()
    for sec in cell.allseclist:
        for seg in sec:
            seg.e_pas = seg.v
            if neuron.h.ismembrane("na_ion", sec=sec):
                seg.e_pas += seg.ina / seg.g_pas
            if neuron.h.ismembrane("k_ion", sec=sec):
                seg.e_pas += seg.ik / seg.g_pas
            if neuron.h.ismembrane("ca_ion", sec=sec):
                seg.e_pas += seg.ica / seg.g_pas
            if neuron.h.ismembrane("Ih", sec=sec):
                seg.e_pas += seg.ihcn_Ih / seg.g_pas
            if neuron.h.ismembrane("Ih_z", sec=sec):
                seg.e_pas += seg.ih_Ih_z / seg.g_pas
            if neuron.h.ismembrane("Ih_linearized_v2_frozen", sec=sec):
                seg.e_pas += seg.ihcn_Ih_linearized_v2_frozen / seg.g_pas
            if neuron.h.ismembrane("Ih_linearized_v2", sec=sec):
                seg.e_pas += seg.ihcn_Ih_linearized_v2 / seg.g_pas
            if neuron.h.ismembrane("Ih_frozen", sec=sec):
                seg.e_pas += seg.ihcn_Ih_frozen / seg.g_pas

def compute_mean_nu_X( params, OUTPUTPATH, TRANSIENT=200.):
    """
    Return the population-averaged firing rate for each population X

    Parameters
    ----------
    params: module
        `<module 'example_network_parameters'>`
    OUTPUTPATH: str
        path to directory with `spikes.h5` file
    TRANSIENT: float
        startup transient duration

    Returns
    -------
    nu_X: dict of floats
        keys and values denote population names and firing rates, respectively.
    """

    import h5py

    nu_X = dict()
    for i, (X, N_X) in enumerate(zip(params.population_names,
                                     params.population_sizes)):
        with h5py.File(os.path.join(OUTPUTPATH, 'spikes.h5'), 'r') as f:
            times = np.concatenate(f[X]['times'][()])
            times = times[times >= TRANSIENT]
            nu_X[X] = (times.size / N_X
                       / (params.networkParameters['tstop'] - TRANSIENT)
                       * 1000)
    return nu_X




import importlib.util
import importlib
from typing import Any, Optional
import subprocess
import sys
import os

def ensure_module(module_name, package_name=None):
    """
    Check if a module is installed; if not, try to install it.

    Args:
        module_name (str): The module to check (e.g., 'rpy2').
        package_name (str, optional): The package name to install (if different from module_name).

    Returns:
        bool: True if the module is installed successfully, False otherwise.
    """
    try:
        if importlib.util.find_spec(module_name) is None:
            print(f"Module '{module_name}' not found. Attempting to install...")
            package = package_name if package_name else module_name
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
            print(f"Module '{module_name}' installed successfully.")

            # Check again after installation
            if importlib.util.find_spec(module_name) is None:
                raise ImportError(f"Installation failed: '{module_name}' not found after installation.")

        return True  # Module is available
    except subprocess.CalledProcessError:
        print(f"Error: Failed to install '{module_name}'. Please install it manually.")
    except ImportError as e:
        print(f"Error: {e}")
    except Exception as e:
        print(f"Unexpected error: {e}")

    return False  # Module is not available


def dynamic_import(
        module_path: str,
        attribute_name: Optional[str] = None,
        package: Optional[str] = None
) -> Any:
    """
    Dynamically import a module or module attribute.

    Args:
        module_path: Full path to the module (e.g., 'package.subpackage.module')
        attribute_name: Optional name of attribute to import from the module
        package: Optional package name for relative imports

    Returns:
        The imported module or attribute

    Raises:
        ImportError: If the module or attribute cannot be imported
    """
    try:
        module = importlib.import_module(module_path, package=package)

        if attribute_name is not None:
            return getattr(module, attribute_name)
        return module

    except ImportError as e:
        raise ImportError(f"Failed to import {module_path}" +
                          (f".{attribute_name}" if attribute_name else "")) from e


def download_zenodo_record(api_url, download_dir="zenodo_files", extract_tar=True, delete_tar=True):
    """
    Download files from a Zenodo record.

    Args:
        api_url (str): The API URL of the Zenodo record.
        download_dir (str): Directory to save downloaded files. Default is 'zenodo_files'.
        extract_tar (bool): Whether to extract tar files after downloading. Default is True.
        delete_tar (bool): Whether to delete tar files after extraction. Default is True.
    """

    # Check if the download directory exists, if not create it
    if not os.path.exists(download_dir):
        os.makedirs(download_dir, exist_ok=True)

        # Check if the required modules are installed
        if not ensure_module("requests"):
            raise ImportError("requests is required for downloading data. ")
        requests = dynamic_import("requests")

        if not ensure_module("tqdm"):
            raise ImportError("tqdm is required for progress bars. ")
        tqdm = dynamic_import("tqdm", "tqdm")

        if not ensure_module("tarfile"):
            raise ImportError("tarfile is required for extracting tar files. ")
        tarfile = dynamic_import("tarfile")

        # Download the Zenodo record
        headers = {"User-Agent": "Mozilla/5.0"}
        response = requests.get(api_url, headers=headers)
        response.raise_for_status()
        record = response.json()

        for file_info in record.get("files", []):
            file_url = file_info["links"]["self"]
            file_name = file_info["key"]
            file_path = os.path.join(download_dir, file_name)

            print(f"Downloading {file_name}...")
            with requests.get(file_url, stream=True) as r:
                r.raise_for_status()
                total_size = int(r.headers.get("content-length", 0))
                with open(file_path, "wb") as f, tqdm(
                    total=total_size,
                    unit='B',
                    unit_scale=True,
                    desc=file_name,
                    initial=0
                ) as progress:
                    for chunk in r.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
                            progress.update(len(chunk))
            print(f"Saved to {file_path}")

        # After all files are downloaded, untar and delete tar files
        if extract_tar:
            for file_name in os.listdir(download_dir):
                file_path = os.path.join(download_dir, file_name)

                if tarfile.is_tarfile(file_path):
                    print(f"Extracting {file_name}...")
                    with tarfile.open(file_path, "r:*") as tar:
                        tar.extractall(path=download_dir)
                    print(f"Extracted to {download_dir}")

                    if delete_tar:
                        os.remove(file_path)
                        print(f"Deleted {file_name}")

    else:
        print(f"Directory {download_dir} already exists. Skipping download.")
        print("If you want to re-download, please delete the directory.")


import neuron
import numpy as np
import scipy.stats as st
from LFPy import NetworkCell

"""
The following methods and parameters were downloaded from the following LFPykernels repositories: 
https://github.com/LFPy/LFPykernels/blob/main/examples/Hagen_et_al_2022_PLOS_Comput_Biol/example_network_parameters.py
https://github.com/LFPy/LFPykernels/blob/main/examples/Hagen_et_al_2022_PLOS_Comput_Biol/example_network_methods.py
Copyright (C) 2021 https://github.com/espenhgn
"""


def set_active(cell, Vrest):
    """Insert passive leak, Ih, NaTa_t and SKv3_1 channels as in Hay 2011 model

    Parameters
    ----------
    cell: object
        LFPy.NetworkCell like object
    Vrest: float
        Steady state potential (not used)
    """
    for sec in cell.template.all:
        sec.insert('pas')
        sec.insert('Ih')
        sec.insert('NaTa_t')
        sec.insert('SKv3_1')
        sec.ena = 50
        sec.ek = -85
        if sec.name().rfind('soma') >= 0:
            sec.gNaTa_tbar_NaTa_t = 2.04
            sec.gSKv3_1bar_SKv3_1 = 0.693
            sec.g_pas = 0.0000338
            sec.e_pas = -90
            sec.gIhbar_Ih = 0.0002

        if sec.name().rfind('apic') >= 0 or sec.name().rfind('dend') >= 0:
            sec.gNaTa_tbar_NaTa_t = 0.0213
            sec.gSKv3_1bar_SKv3_1 = 0.000261
            sec.g_pas = 0.0000589
            sec.e_pas = -90
            sec.gIhbar_Ih = 0.0002 * 10

class KernelParams:
    # Transient time
    transient = 2000

    # Time lag relative to spike for kernel predictions
    tau = 100

    # Parameters of connectivity and synaptic weights
    MC_params = {
        'weight_EE': 0.00015,
        'weight_IE': 0.000125,
        'weight_EI': 0.0045,
        'weight_II': 0.002,
        'weight_scaling': 1.0,
        'biophys': 'lin',
        'i_syn': True,
        'n_ext': [465, 160],
        'g_eff': True,
        'perseg_Vrest': False}

    # testing flag. If True, run with downsized network w. dense connectivity
    TESTING = False

    # class NetworkCell parameters:
    cellParameters = dict(
        # morphology='BallAndStick.hoc',  # set by main simulation
        templatefile='BallAndSticksTemplate.hoc',
        templatename='BallAndSticksTemplate',
        custom_fun=[set_active],
        custom_fun_args=[dict(Vrest=-65.)],  # [dict(Vrest=Vrest)] set at runtime
        templateargs=None,
        delete_sections=False,
    )

    # class NetworkPopulation parameters:
    populationParameters = dict(
        Cell=NetworkCell,
        cell_args=cellParameters,
        pop_args=dict(
            radius=150.,  # population radius
            loc=0.,  # population center along z-axis
            scale=75.),  # standard deviation along z-axis
        rotation_args=dict(x=0., y=0.))

    # class Network parameters:
    networkParameters = dict(
        dt=2 ** -4,  # simulation time resolution (ms)
        tstop=12000.,  # simulation duration (ms)
        v_init=-65.,  # initial membrane voltage for all cells (mV)
        celsius=34,  # simulation temperature (deg. C)
        # OUTPUTPATH=OUTPUTPATH  # set in main simulation script
    )

    # class RecExtElectrode parameters:
    electrodeParameters = dict(
        x=np.zeros(13),  # x-coordinates of contacts
        y=np.zeros(13),  # y-coordinates of contacts
        z=np.linspace(1000., -200., 13),  # z-coordinates of contacts
        N=np.array([[0., 1., 0.] for _ in range(13)]),  # contact surface normals
        r=5.,  # contact radius
        n=100,  # n-point averaging for potential on each contact
        sigma=0.3,  # extracellular conductivity (S/m)
        method="linesource"  # use line sources
    )

    # class LaminarCurrentSourceDensity parameters:
    csdParameters = dict(
        z=np.c_[electrodeParameters['z'] - 50.,
                electrodeParameters['z'] + 50.],  # lower and upper boundaries
        r=np.array([populationParameters['pop_args']['radius']] * 13)  # radius
    )

    # method Network.simulate() parameters:
    networkSimulationArguments = dict(
        rec_pop_contributions=True,  # store contributions by each population
        to_memory=True,  # simulate to memory
        to_file=False  # simulate to file
    )

    # population names, morphologies, sizes and connection probability:
    population_names = ['E', 'I']
    morphologies = ['BallAndSticks_E.hoc', 'BallAndSticks_I.hoc']
    if TESTING:
        population_sizes = [32, 8]
        connectionProbability = [[1., 1.], [1., 1.]]
    else:
        population_sizes = [8192, 1024]
        connectionProbability = [[0.05, 0.05], [0.05, 0.05]]

    # synapse model. All corresponding parameters for weights,
    # connection delays, multapses and layerwise positions are
    # set up as shape (2, 2) nested lists for each possible
    # connection on the form:
    # [["E:E", "E:I"],
    #  ["I:E", "I:I"]].
    # using convention "<pre>:<post>"
    synapseModel = neuron.h.Exp2Syn
    # synapse parameters in terms of rise- and decay time constants
    # (tau1, tau2 [ms]) and reversal potential (e [mV])
    synapseParameters = [[dict(tau1=0.2, tau2=1.8, e=0.),
                          dict(tau1=0.2, tau2=1.8, e=0.)],
                         [dict(tau1=0.1, tau2=9.0, e=-80.),
                          dict(tau1=0.1, tau2=9.0, e=-80.)]]
    # synapse max. conductance (function, mean, st.dev., min.):
    weightFunction = np.random.normal
    # weight_<pre><post> values set by parameters file via main simulation scripts
    # weightArguments = [[dict(loc=weight_EE, scale=weight_EE / 10),
    #                     dict(loc=weight_IE, scale=weight_IE / 10)],
    #                    [dict(loc=weight_EI, scale=weight_EI / 10),
    #                     dict(loc=weight_II, scale=weight_II / 10)]]
    minweight = 0.  # weight values below this value will be redrawn

    # conduction delay (function, mean, st.dev., min.) using truncated normal
    # continuous random variable:
    delayFunction = st.truncnorm
    delayArguments = [[dict(a=(0.3 - 1.5) / 0.3, b=np.inf, loc=1.5, scale=0.3),
                       dict(a=(0.3 - 1.4) / 0.4, b=np.inf, loc=1.4, scale=0.4)],
                      [dict(a=(0.3 - 1.4) / 0.5, b=np.inf, loc=1.3, scale=0.5),
                       dict(a=(0.3 - 1.2) / 0.6, b=np.inf, loc=1.2, scale=0.6)]]
    # will be deprecated; has no effect with delayFunction = st.truncnorm:
    mindelay = None

    # Distributions of multapses. They are here defined via a truncated normal
    # continous random variable distribution which will be used to compute a
    # discrete probability distribution for integer numbers of
    # synapses 1, 2, ..., 100, via a scipy.stats.rv_discrete instance
    multapseFunction = st.truncnorm
    multapseArguments = [[dict(a=(1 - 2.) / .4,
                               b=(10 - 2.) / .4,
                               loc=2.,
                               scale=.4),
                          dict(a=(1 - 2.) / .6,
                               b=(10 - 2.) / .6,
                               loc=2.,
                               scale=.6)],
                         [dict(a=(1 - 5.) / 0.9,
                               b=(10 - 5.) / 0.9,
                               loc=5.,
                               scale=0.9),
                          dict(a=(1 - 5.) / 1.1,
                               b=(10 - 5.) / 1.1,
                               loc=5.,
                               scale=1.1)]]

    # method NetworkCell.get_rand_idx_area_and_distribution_norm
    # parameters for layerwise synapse positions:
    synapsePositionArguments = [[dict(section=['apic', 'dend'],
                                      fun=[st.norm, st.norm],
                                      funargs=[dict(loc=0., scale=100.),
                                               dict(loc=500., scale=100.)],
                                      funweights=[0.5, 1.]
                                      ),
                                 dict(section=['apic', 'dend'],
                                      fun=[st.norm],
                                      funargs=[dict(loc=50., scale=100.)],
                                      funweights=[1.]
                                      )],
                                [dict(section=['soma', 'apic', 'dend'],
                                      fun=[st.norm],
                                      funargs=[dict(loc=-50., scale=100.)],
                                      funweights=[1.]
                                      ),
                                 dict(section=['soma', 'apic', 'dend'],
                                      fun=[st.norm],
                                      funargs=[dict(loc=-100., scale=100.)],
                                      funweights=[1.]
                                      )]
                                ]

    # Parameters for extrinsic (e.g., cortico-cortical connections) synapses
    # and mean interval (ms)
    extSynapseParameters = dict(
        syntype='Exp2Syn',
        weight=0.0002,
        tau1=0.2,
        tau2=1.8,
        e=0.
    )
    netstim_interval = 25.


import sys
import os
import pickle
import pandas as pd
import scipy.signal as ss
import numpy as np
import time
from matplotlib import pyplot as plt
import ncpi
from ncpi import tools

# Path to parameters of the LIF network model
sys.path.append(os.path.join(os.path.dirname(__file__), '../simulation/params'))

# Choose to either download files and precomputed outputs used in simulations of the reference multicompartment neuron
# network model (True) or load them from a local path (False)
zenodo_dw_mult = True

# Zenodo URL that contains the data (used if zenodo_dw_mult is True)
zenodo_URL_mult = "https://zenodo.org/api/records/15429373"

# Zenodo directory where the data is stored (must be an absolute path to correctly load morphologies in NEURON)
zenodo_dir = '/DATA/multicompartment_neuron_network'

# Set to True to run new simulations of the LIF network model, or False to load precomputed results from a pickle file
# located in a 'data' folder.
compute_new_sim = True

# Number of repetitions of each simulation
trials = 6

# Configurations of parameters to simulate:
# Best fit
# confs = [[1.589, 2.020, -23.84, -8.441, 0.5, 0.5, 29.89]]

# Changing J_ext
confs = [[1.589, 2.020, -23.84, -8.441, 0.5, 0.5, 28.],
          [1.589, 2.020, -23.84, -8.441, 0.5, 0.5, 30.],
          [1.589, 2.020, -23.84, -8.441, 0.5, 0.5, 32.]]

# Do not change these paths if the zenodo_dir has been correctly set:
# (1) Simulation output from the multicompartment neuron network model
output_path = os.path.join(zenodo_dir, 'multicompartment_neuron_network', 'output', 'adb947bfb931a5a8d09ad078a6d256b0')

# (2) Path to the data files of the multicompartment neuron models
multicompartment_neuron_network_path = os.path.join(zenodo_dir, 'multicompartment_neuron_network')

# Names of catch22 features
try:
    import pycatch22
    catch22_names = pycatch22.catch22_all([0])['names']
except:
    catch22_names = ['DN_HistogramMode_5',
                     'DN_HistogramMode_10',
                     'CO_f1ecac',
                     'CO_FirstMin_ac',
                     'CO_HistogramAMI_even_2_5',
                     'CO_trev_1_num',
                     'MD_hrv_classic_pnn40',
                     'SB_BinaryStats_mean_longstretch1',
                     'SB_TransitionMatrix_3ac_sumdiagcov',
                     'PD_PeriodicityWang_th0_01',
                     'CO_Embed2_Dist_tau_d_expfit_meandiff',
                     'IN_AutoMutualInfoStats_40_gaussian_fmmi',
                     'FC_LocalSimple_mean1_tauresrat',
                     'DN_OutlierInclude_p_001_mdrmd',
                     'DN_OutlierInclude_n_001_mdrmd',
                     'SP_Summaries_welch_rect_area_5_1',
                     'SB_BinaryStats_diff_longstretch0',
                     'SB_MotifThree_quantile_hh',
                     'SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1',
                     'SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1',
                     'SP_Summaries_welch_rect_centroid',
                     'FC_LocalSimple_mean3_stderr']

def get_spike_rate(times, transient, dt, tstop):
    """
    Compute the spike rate from spike times.

    Parameters
    ----------
    times : array
        Spike times.
    transient : float
        Transient time at the start of the simulation.
    dt : float
        Simulation time step or bin size.
    tstop : float
        Simulation stop time.

    Returns
    -------
    bins : array
        Time bins.
    hist : array
        Spike rate.
    """
    bins = np.arange(transient, tstop + dt, dt)
    hist, _ = np.histogram(times, bins=bins)
    return bins, hist.astype(float)

# Download data
if zenodo_dw_mult:
    print('\n--- Downloading data.')
    start_time = time.time()
    tools.download_zenodo_record(zenodo_URL_mult, download_dir=zenodo_dir)
    end_time = time.time()
    print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")

# Random seed for numpy
np.random.seed(0)

# Simulation outputs
spikes = [[] for _ in range(trials)]
CDMs = [[] for _ in range(trials)]

for trial in range(trials):
    for k,params in enumerate(confs):
        if compute_new_sim:
            print(f'\nTrial {trial+1}/{trials}, Configuration {k+1}/{len(confs)}')
            # Parameters of the model
            J_EE = params[0]
            J_IE = params[1]
            J_EI = params[2]
            J_II = params[3]
            tau_syn_E = params[4]
            tau_syn_I = params[5]
            J_ext = params[6]

            # Load LIF_params
            from network_params import LIF_params

            # Modify parameters
            LIF_params['J_YX'] = [[J_EE, J_IE], [J_EI, J_II]]
            LIF_params['tau_syn_YX'] = [[tau_syn_E, tau_syn_I],
                                        [tau_syn_E, tau_syn_I]]
            LIF_params['J_ext'] = J_ext

            # Create a Simulation object
            sim = ncpi.Simulation(param_folder='../simulation/params',
                                  python_folder='../simulation/python',
                                  output_folder='../simulation/output')

            # Save parameters to a pickle file
            with open(os.path.join('../simulation/output', 'network.pkl'), 'wb') as f:
                pickle.dump(LIF_params, f)

            # Run the simulation
            sim.simulate('simulation.py', 'simulation_params.py')

            # Load spike times
            with open(os.path.join('../simulation/output', 'times.pkl'), 'rb') as f:
                times = pickle.load(f)

            # Load gids
            with open(os.path.join('../simulation/output', 'gids.pkl'), 'rb') as f:
                gids = pickle.load(f)

            # Load tstop
            with open(os.path.join('../simulation/output', 'tstop.pkl'), 'rb') as f:
                tstop = pickle.load(f)

            # Load dt
            with open(os.path.join('../simulation/output', 'dt.pkl'), 'rb') as f:
                dt = pickle.load(f)

            # Load X and N_X
            with open(os.path.join('../simulation/output', 'network.pkl'), 'rb') as f:
                LIF_params = pickle.load(f)
                P_X = LIF_params['X']
                N_X = LIF_params['N_X']

            # Transient period
            from analysis_params import KernelParams
            transient = KernelParams.transient
            for X in P_X:
                gids[X] = gids[X][times[X] >= transient]
                times[X] = times[X][times[X] >= transient]

            # Compute the kernel
            print('Computing the kernel...')
            potential = ncpi.FieldPotential()
            biophys = ['set_Ih_linearized_hay2011', 'make_cell_uniform']

            H_YX = potential.create_kernel(multicompartment_neuron_network_path,
                                           output_path,
                                           KernelParams,
                                           biophys,
                                           dt,
                                           tstop,
                                           electrodeParameters=None,
                                           CDM=True)

            # Compute CDM
            probe = 'KernelApproxCurrentDipoleMoment'
            CDM_data = dict(EE=[], EI=[], IE=[], II=[])

            for X in P_X:
                for Y in P_X:
                    # Compute the firing rate
                    bins, spike_rate = get_spike_rate(times[X], transient, dt, tstop)
                    # Pick only the z-component of the CDM kernel
                    kernel = H_YX[f'{X}:{Y}'][probe][2, :]
                    # CDM
                    sig = np.convolve(spike_rate, kernel, 'same')
                    CDM_data[f'{X}{Y}'] = ss.decimate(sig,
                                                      q=10,
                                                      zero_phase=True)

            # Collect the simulation outputs
            spikes[trial].append([times, gids])
            CDMs[trial].append(CDM_data)

            # Save the simulation outputs to a pickle file
            if not os.path.exists('data'):
                os.makedirs('data')
            with open(f'data/output_{k}_{trial}.pkl', 'wb') as f:
                pickle.dump([times, gids, CDM_data, dt, tstop, transient, P_X, N_X], f)

        else:
            try:
                with open(f'data/output_{k}_{trial}.pkl', 'rb') as f:
                    times, gids, CDM_data, dt, tstop, transient, P_X, N_X = pickle.load(f)
                spikes[trial].append([times, gids])
                CDMs[trial].append(CDM_data)
            except FileNotFoundError:
                print(f'File data/output_{k}_{trial}.pkl not found. Please run the simulation first.')
                sys.exit(1)


# Create a figure and set its properties
fig = plt.figure(figsize=(7.5, 6.), dpi=300)
plt.rcParams.update({'font.size': 10, 'font.family': 'Arial'})
plt.rc('xtick', labelsize=8)
plt.rc('ytick', labelsize=8)

# Time interval
T = [4000, 4100]

# Raster plot of the spike trains
colors = ['#1f77b4', '#ff7f0e']
for col in range(3):
    ax = fig.add_axes([0.1 + col * 0.3, 0.73, 0.25, 0.22])
    for i,X in enumerate(P_X):
        t = spikes[0][col][0][X] # pick the first trial
        gi = spikes[0][col][1][X]
        # gi = gi[t >= transient]
        # t = t[t >= transient]

        # Spikes
        ii = (t >= T[0]) & (t <= T[1])
        ax.plot(t[ii], gi[ii], '.', color = colors[i], markersize=0.5)

    ax.set_title(r'$J_{syn}^{ext}$ = %s nA' % confs[col][6])
    if col == 0:
        ax.set_ylabel('Neuron ID')
        ax.yaxis.set_label_coords(-0.22, 0.5)

        # Fake legend
        for j, Y in enumerate(P_X):
            ax.plot([], [], '.', color = colors[j], label=f'{Y}', markersize=4)
        ax.legend(loc=1, fontsize=8, labelspacing=0.2)
    else:
        ax.set_yticks([])
        ax.set_yticklabels([])
        ax.set_ylabel('')
    ax.axis('tight')
    ax.set_xticklabels([])
    ax.set_xticks([])


# Firing rates
for col in range(3):
    ax = fig.add_axes([0.1 + col * 0.3, 0.6, 0.25, 0.12])
    for i,X in enumerate(P_X):
        # Compute the firing rate
        bins, spike_rate = get_spike_rate(spikes[0][col][0][X], transient, dt, tstop)
        # Plot the firing rate
        bins = bins[:-1]
        ii = (bins >= T[0]) & (bins <= T[1])
        ax.plot(bins[ii], spike_rate[ii], color='C{}'.format(i),label=r'$\nu_\mathrm{%s}$' % X)

    if col == 0:
        ax.legend(loc=1)
        ax.set_ylabel(r'$\nu_X$ (spik./$\Delta t$)')
        ax.yaxis.set_label_coords(-0.22, 0.5)

    ax.axis('tight')
    ax.set_xticklabels([])
    ax.set_xticks([])

# CDMs
for col in range(3):
    ax = fig.add_axes([0.1 + col * 0.3, 0.47, 0.25, 0.12])
    CDM = CDMs[0][col]['EE'] + CDMs[0][col]['EI'] + CDMs[0][col]['IE'] + CDMs[0][col]['II']
    bins = np.arange(transient, tstop, dt)
    bins = bins[::10]  # to take into account the decimate ratio
    ii = (bins >= T[0]) & (bins <= T[1])
    ax.plot(bins[ii], CDM[ii], color='k')

    if col == 0:
        ax.set_ylabel(r'CDM ($P_z$)')
        ax.yaxis.set_label_coords(-0.22, 0.5)
    ax.set_yticks([])
    ax.set_xlabel('t (ms)')
    ax.axis('tight')

    # Add scale
    y_max = np.max(CDM[ii])
    y_min = np.min(CDM[ii])
    scale = (y_max - y_min) / 5
    ax.plot([T[0] if col < 2 else T[0] + 50,T[0] if col < 2 else T[0] + 50],
             [y_min + scale, y_min], 'k')
    ax.text(T[0] + 1 if col < 2 else T[0] + 51,
            y_min + scale/4., r'$2^{%s}nAcm$' % np.round(np.log2(scale*10**(-4))), fontsize=8)

# Power spectra
ax = fig.add_axes([0.1, 0.07, 0.27, 0.3])
colors = ['C0', 'C1', 'C2']
for col in range(3):
    CDM = [CDMs[trial][col]['EE'] + CDMs[trial][col]['EI'] +
           CDMs[trial][col]['IE'] + CDMs[trial][col]['II'] for trial in range(trials)]
    f, Pxx = ss.welch(CDM, fs=1000./(10.*dt))
    # Trial-averaged power spectrum
    Pxx = np.mean(Pxx, axis=0)
    # Normalize the power spectrum
    Pxx = Pxx / np.sum(Pxx)
    f1 = f[f >= 10]
    f2 = f1[f1 <= 200]
    ax.semilogy(f2, Pxx[(f >= 10) & (f <= 200)], label=r'$J_{syn}^{ext}$ = %s nA' % confs[col][6],
                color=colors[col])
ax.legend(loc='lower left', fontsize=8, labelspacing=0.2)
ax.set_xlabel('Frequency (Hz)')
ax.set_ylabel('Normalized power')

# Collect the CDMs
all_CDMs = []
IDs = []
epochs = []
for trial in range(trials):
    for k,params in enumerate(confs):
        all_CDMs.append(CDMs[trial][k]['EE'] + CDMs[trial][k]['EI'] + CDMs[trial][k]['IE'] + CDMs[trial][k]['II'])
        IDs.append(k)
        epochs.append(trial)

# Compute features
all_features = {}
all_methods = ['catch22', 'power_spectrum_parameterization']
for method in all_methods:
    print(f'\n\n--- Method: {method}')

    # Create a fake Pandas DataFrame (only Data and fs are relevant)
    df = pd.DataFrame({'ID': IDs,
                       'Group': IDs,
                       'Epoch': epochs,
                       'Sensor': np.zeros(len(IDs)),  # dummy sensor
                       'Data': all_CDMs})
    df.Recording = 'LFP'
    df.fs = 1000. / (10. * dt)

    # Compute features
    if method == 'catch22':
        features = ncpi.Features(method='catch22')
    elif method == 'power_spectrum_parameterization':
        # Parameters of the fooof algorithm
        fooof_setup_sim = {'peak_threshold': 1.,
                           'min_peak_height': 0.,
                           'max_n_peaks': 5,
                           'peak_width_limits': (10., 50.)}
        features = ncpi.Features(method='power_spectrum_parameterization',
                                 params={'fs': df.fs,
                                         'fmin': 5.,
                                         'fmax': 200.,
                                         'fooof_setup': fooof_setup_sim,
                                         'r_squared_th': 0.9})

    df = features.compute_features(df)

    # Keep only the aperiodic exponent
    if method == 'power_spectrum_parameterization':
        df['Features'] = df['Features'].apply(lambda x: x[1])

    # Append the feature dataframes to a list
    all_features[method] = df

# Plot features
colors = ['lightcoral', 'lightblue', 'lightgreen', 'lightgrey']
for row in range(2):
    for col in range(2):
        ax = fig.add_axes([0.5 + col * 0.27, 0.24 - row * 0.16, 0.18, 0.13])

        if row == 0 and col == 0:
            feats = np.array(all_features['power_spectrum_parameterization']['Features'].tolist())
            ax.set_ylabel(r'$1/f$' + ' ' + r'$slope$')
        if row == 0 and col == 1:
            feats = np.array(all_features['catch22']['Features'].tolist())
            idx = catch22_names.index('SC_FluctAnal_2_dfa_50_1_2_logi_prop_r1')
            feats = feats[:, idx]
            ax.set_ylabel(r'$dfa$')
        if row == 1 and col == 0:
            feats = np.array(all_features['catch22']['Features'].tolist())
            idx = catch22_names.index('SC_FluctAnal_2_rsrangefit_50_1_logi_prop_r1')
            feats = feats[:, idx]
            ax.set_ylabel(r'$rs\ range$')
        if row == 1 and col == 1:
            feats = np.array(all_features['catch22']['Features'].tolist())
            idx = catch22_names.index('MD_hrv_classic_pnn40')
            feats = feats[:, idx]
            ax.set_ylabel(r'$high\ fluct.$')

        # Rearrange the features
        feats_plot = np.zeros((trials,len(confs)))
        for conf in range(len(confs)):
            feats_plot[:,conf] = feats[np.array(IDs) == conf]

        # Plot the average values
        ax.plot(np.arange(len(confs)), np.mean(feats_plot, axis=0), color=colors[row*2 + col])

        # Plot the variance
        ax.fill_between(np.arange(len(confs)), np.mean(feats_plot, axis=0) - np.std(feats_plot, axis=0),
                        np.mean(feats_plot, axis=0) + np.std(feats_plot, axis=0), color=colors[row*2 + col],
                        alpha=0.3)

        # Labels
        if row == 1:
            ax.set_xlabel(r'$J_{syn}^{ext}$ (nA)')
            ax.set_xticks(np.arange(3))
            ax.set_xticklabels([f'{confs[i][6]}' for i in range(3)])
        else:
            ax.set_xticks([])
            ax.set_xticklabels([])

# Plot letters
ax = fig.add_axes([0., 0., 1., 1.])
ax.axis('off')
ax.text(0.01, 0.97, 'A', fontsize=12, fontweight='bold')
ax.text(0.01, 0.37, 'B', fontsize=12, fontweight='bold')
ax.text(0.4, 0.37, 'C', fontsize=12, fontweight='bold')

# Save the figure
plt.savefig('example_full_pipeline.png', bbox_inches='tight')
# plt.show()



import os
import pickle
import numpy as np
from pathos.multiprocessing import ProcessingPool as Pool
from tqdm import tqdm

# Path to the folders containing the simulation data
sim_file_path_pre = '/DATOS/pablomc/Hagen_model_v1'

# Path to the folder containing the processed data
sim_file_path_post = '/DATOS/pablomc/data/Hagen_model_v1'


def process_batch(ldir):
    """
    Load simulation data and create the data structures of normalized CDMs and
    LIF network model parameters.

    Parameters
    ----------
    ldir: list of string
        Absolute paths of the folders that contain the simulation data.

    Returns
    -------
    CDM_data: nested list
        Normalized CDMs.
    theta_data: nested list
        Synapse parameters of the LIF network model.
    """
    theta_data = []
    CDM_data = []

    for folder in ldir:
        # Load and sum CDMs for all combinations of populations (EE, EI, IE, II)
        # if CDM_data file is a dict
        try:
            cdm = pickle.load(open(os.path.join(folder,"CDM_data"),'rb'))
            if isinstance(cdm, dict):
                cdm_sum = cdm['EE'] + cdm['EI'] + cdm['IE'] + cdm['II']
            else:
                cdm_sum = cdm

            # Dismiss CDMs that are constant over time
            if np.std(cdm_sum) > 10 ** (-10):
                # Remove the first 500 samples containing the transient response (not really necessary as some
                # transient time is already removed when computing the CDMs)
                cdm_sum = cdm_sum[500:]
                # Normalization
                CDM_data.append((cdm_sum - np.mean(cdm_sum)) / np.std(cdm_sum))
                # Collect synapse parameters of recurrent connections and
                # external input
                try:
                    LIF_params = pickle.load(open(os.path.join(
                        folder, "LIF_params"), 'rb'))
                    theta_data.append([LIF_params['J_YX'][0][0],
                                       LIF_params['J_YX'][0][1],
                                       LIF_params['J_YX'][1][0],
                                       LIF_params['J_YX'][1][1],
                                       LIF_params['tau_syn_YX'][0][0],
                                       LIF_params['tau_syn_YX'][0][1],
                                       LIF_params['J_ext']
                                       ])

                except (FileNotFoundError, IOError):
                    print(f'File LIF_params not found in {folder}')

        except (FileNotFoundError, IOError):
            print(f'File CDM_data not found in {folder}')

    return CDM_data, theta_data


if __name__ == '__main__':

    # List of all the folders containing the simulation data (there are three folders that correspond
    # to the different computing environments used to run the simulations)
    folder1 = os.path.join(sim_file_path_pre, 'LIF_simulations')
    folder2 = os.path.join(sim_file_path_pre, 'LIF_simulations_hpmoon','LIF_simulations')
    folder3 = os.path.join(sim_file_path_pre, 'LIF_simulations_hpc','LIF_simulations')

    ldir = [os.path.join(folder1, f) for f in os.listdir(folder1)] + \
           [os.path.join(folder2, f) for f in os.listdir(folder2)] + \
           [os.path.join(folder3, f) for f in os.listdir(folder3)]

    # Dictionary to store parameters of the LIF network model
    theta_data = {'parameters':['J_EE',
                                'J_IE',
                                'J_EI',
                                'J_II',
                                'tau_syn_E',
                                'tau_syn_I',
                                'J_ext'],
                  'data': []}
    # Current Dipole Moment (CDM) data
    CDM_data = []

    # Split the list of folders into batches
    batch_size_1 = len(ldir) // 5 # 5 is a factor to avoid memory issues
                                  # (set it to a smaller size if your system has sufficient memory)
    batches_1 = [ldir[i:i + batch_size_1] for i in range(0, len(ldir), batch_size_1)]

    # Preprocess data in parallel using all available CPUs
    num_cpus = os.cpu_count()
    for ii, batch in enumerate(batches_1):
        print(f"Processing batch {ii+1}/{len(batches_1)}")
        batch_size_2 = len(batch) // (num_cpus*10) # 10 is a factor to update the progress bar more frequently
        if batch_size_2 < 1:
            batch_size_2 = 1
        batches_2 = [batch[i:i + batch_size_2] for i in range(0, len(batch), batch_size_2)]

        with Pool(num_cpus) as pool:
            results = list(tqdm(pool.imap(process_batch, batches_2),
                                total=len(batches_2), desc="Processing data"))

        # Collect the results
        for result in results:
            CDM_data.extend(result[0])
            theta_data['data'].extend(result[1])

        # Transform to numpy arrays
        theta_data['data'] = np.array(theta_data['data'],dtype="float32")
        CDM_data = np.array(CDM_data,dtype="float32")

        print(f"Number of simulations in the batch: {len(batch)}")
        print(f"Number of samples processed: {CDM_data.shape[0]}\n")

        # Create folders if they do not exist
        splits = os.path.split(sim_file_path_post)
        if not os.path.isdir(splits[0]):
            os.mkdir(splits[0])
        if not os.path.isdir(sim_file_path_post):
            os.mkdir(sim_file_path_post)

        # Save numpy arrays to file
        pickle.dump(theta_data,open(os.path.join(sim_file_path_post,f'theta_data_{ii}'),'wb'))
        pickle.dump(CDM_data,open(os.path.join(sim_file_path_post,f'CDM_data_{ii}'),'wb'))

        # Clear memory
        theta_data['data'] = []
        CDM_data = []
        results = []



import os
import shutil
import pickle
import pandas as pd
import scipy
import numpy as np
import time
import ncpi
from ncpi import tools

# Choose to either download data from Zenodo (True) or load it from a local path (False).
# Important: the zenodo downloads will take a while, so if you have already downloaded the data, set this to False and
# configure the zenodo_dir variables to point to the local paths where the data is stored.
zenodo_dw_sim = True # simulation data
zenodo_dw_emp = True # empirical data

# Zenodo URL that contains the simulation data and ML models (used if zenodo_dw_sim is True)
zenodo_URL_sim = "https://zenodo.org/api/records/15351118"

# Zenodo URL that contains the empirical data (used if zenodo_dw_emp is True)
zenodo_URL_emp = "https://zenodo.org/api/records/15382047"

# Paths to zenodo files
zenodo_dir_sim = "zenodo_sim_files"
zenodo_dir_emp= "zenodo_emp_files"

# Methods used to compute the features
all_methods = ['catch22','power_spectrum_parameterization_1']

# ML model used to compute the predictions (MLPRegressor, Ridge or NPE)
ML_model = 'MLPRegressor'

if __name__ == "__main__":
    # Download simulation data and ML models
    if zenodo_dw_sim:
        print('\n--- Downloading simulation data and ML models from Zenodo.')
        start_time = time.time()
        tools.download_zenodo_record(zenodo_URL_sim, download_dir=zenodo_dir_sim)
        end_time = time.time()
        print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")

    # Download empirical data
    if zenodo_dw_emp:
        print('\n--- Downloading empirical data from Zenodo.')
        start_time = time.time()
        tools.download_zenodo_record(zenodo_URL_emp, download_dir=zenodo_dir_emp)
        end_time = time.time()
        print(f"All files downloaded in {(end_time - start_time) / 60:.2f} minutes.")

    # Iterate over the methods used to compute the features
    for method in all_methods:
        print(f'\n\n--- Method: {method}')

        # Load parameters of the model (theta) and features (X) from simulation data
        print('\n--- Loading simulation data.')
        start_time = time.time()

        try:
            with open(os.path.join(zenodo_dir_sim, 'data', method, 'sim_theta'), 'rb') as file:
                theta = pickle.load(file)
            with open(os.path.join(zenodo_dir_sim, 'data', method, 'sim_X'), 'rb') as file:
                X = pickle.load(file)
        except Exception as e:
            print(f"Error loading simulation data: {e}")

        # Print info
        print('theta:')
        for key, value in theta.items():
            if isinstance(value, np.ndarray):
                print(f'--Shape of {key}: {value.shape}')
            else:
                print(f'--{key}: {value}')
        print(f'Shape of X: {X.shape}')

        end_time = time.time()
        print(f"Done in {(end_time - start_time):.2f} sec.")

        # Load the Inference objects and add the simulation data
        print('\n--- Loading the inverse model.')
        start_time = time.time()

        # Create inference object
        inference = ncpi.Inference(model=ML_model)
        # Not sure if this is really needed
        inference.add_simulation_data(X, theta['data'])

        # Create folder to save results
        if not os.path.exists('data'):
            os.makedirs('data')
        if not os.path.exists(os.path.join('data', method)):
            os.makedirs(os.path.join('data', method))

        # Transfer model and scaler to the data folder
        if ML_model == 'MLPRegressor':
            folder = 'MLP'
        elif ML_model == 'Ridge':
            folder = 'Ridge'
        elif ML_model == 'NPE':
            folder = 'SBI'
            
        shutil.copy(
            os.path.join(zenodo_dir_sim, 'ML_models/4_param', folder, method, 'scaler'),
            os.path.join('data', 'scaler.pkl')
        )

        shutil.copy(
            os.path.join(zenodo_dir_sim, 'ML_models/4_param', folder, method, 'model'),
            os.path.join('data', 'model.pkl')
        )

        if ML_model == 'NPE':
            shutil.copy(
                os.path.join(zenodo_dir_sim, 'ML_models/4_param', folder, method,
                             'density_estimator'),
                os.path.join('data', 'density_estimator.pkl')
            )

        end_time = time.time()
        print(f'Done in {(end_time - start_time):.2f} sec')

        # Load empirical data
        print('\n--- Loading LFP data.')
        start_time = time.time()

        file_list = os.listdir(os.path.join(zenodo_dir_emp, 'development_EI_decorrelation/baseline/LFP'))
        emp_data = {'LFP': [], 'fs': [], 'age': []}

        for i,file_name in enumerate(file_list):
            print(f'\r Progress: {i+1} of {len(file_list)} files loaded', end='', flush=True)
            structure = scipy.io.loadmat(os.path.join(os.path.join(zenodo_dir_emp,
                                                                   'development_EI_decorrelation/baseline/LFP'),
                                                      file_name))
            LFP = structure['LFP']['LFP'][0,0]
            sum_LFP = np.sum(LFP, axis=0)  # sum LFP across channels
            fs = structure['LFP']['fs'][0, 0][0, 0]
            age = structure['LFP']['age'][0,0][0,0]

            emp_data['LFP'].append(sum_LFP)
            emp_data['fs'].append(fs)
            emp_data['age'].append(age)

        end_time = time.time()
        print(f'\nFiles loaded: {len(emp_data["LFP"])}')
        print(f'Done in {(end_time - start_time):.2f} sec')

        # Compute features from empirical data
        print('\n--- Computing features from empirical data.')
        start_time = time.time()

        # Epoch length in seconds
        chunk_size = 5

        # Parameters of the feature extraction method
        if method == 'catch22':
            params = None
        elif method == 'power_spectrum_parameterization_1':
            fooof_setup_emp = {'peak_threshold': 1.,
                               'min_peak_height': 0.,
                               'max_n_peaks': 5,
                               'peak_width_limits': (10., 50.)}
            params={'fs': emp_data['fs'][0],
                   'fmin': 5.,
                   'fmax': 45.,
                   'fooof_setup': fooof_setup_emp,
                   'r_squared_th':0.9}

        # Split the data into chunks (epochs)
        chunk_size = int(chunk_size * emp_data['fs'][0])
        chunked_data = []
        ID = []
        epoch = []
        group = []
        for i in range(len(emp_data['LFP'])):
            for e,j in enumerate(range(0, len(emp_data['LFP'][i]), chunk_size)):
                if len(emp_data['LFP'][i][j:j+chunk_size]) == chunk_size:
                    chunked_data.append(emp_data['LFP'][i][j:j+chunk_size])
                    ID.append(i)
                    epoch.append(e)
                    group.append(emp_data['age'][i])

        # Create the Pandas DataFrame
        df = pd.DataFrame({'ID': ID,
                           'Group': group,
                           'Epoch': epoch,
                           'Sensor': np.zeros(len(ID)), # dummy sensor
                           'Data': chunked_data})
        df.Recording = 'LFP'
        df.fs = emp_data['fs'][0]

        # Compute features
        features = ncpi.Features(method=method if method == 'catch22' else 'power_spectrum_parameterization',
                                 params=params)
        emp_data = features.compute_features(df)

        # Keep only the aperiodic exponent (1/f slope)
        if method == 'power_spectrum_parameterization_1':
            emp_data['Features'] = emp_data['Features'].apply(lambda x: x[1])

        end_time = time.time()
        print(f'Done in {(end_time - start_time):.2f} sec')

        # Compute predictions from the empirical data
        print('\n--- Computing predictions from empirical data.')
        start_time = time.time()
        # Predict the parameters from the features of the empirical data
        predictions = inference.predict(np.array(emp_data['Features'].tolist()))

        # Append the predictions to the DataFrame
        pd_preds = pd.DataFrame({'Predictions': predictions})
        emp_data = pd.concat([emp_data, pd_preds], axis=1)
        end_time = time.time()
        print(f'Done in {(end_time - start_time)/60.} min')

        # Save the data including predictions of all parameters
        emp_data.to_pickle(os.path.join('data', method, 'emp_data_all.pkl'))

        # Replace parameters of recurrent synaptic conductances with the ratio (E/I)_net
        E_I_net = emp_data['Predictions'].apply(lambda x: (x[0]/x[2]) / (x[1]/x[3]))
        others = emp_data['Predictions'].apply(lambda x: x[4:])
        emp_data['Predictions'] = (np.concatenate((E_I_net.values.reshape(-1,1),
                                                   np.array(others.tolist())), axis=1)).tolist()

        # Save the data including predictions of (E/I)_net
        emp_data.to_pickle(os.path.join('data', method, 'emp_data_reduced.pkl'))



import os
import pickle
import sys
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import ncpi

# Folder with parameters of LIF model simulations
sys.path.append(os.path.join(os.path.dirname(__file__), '../../simulation/Hagen_model/simulation/params'))

# Path to the folder with prediction results
pred_results = '../data'

# Calculate new firing rates (True) or load them from file if they already exist (False). If firing rates do not
# exist, they will not be plotted.
compute_firing_rate = False

# Path to saved firing rates
fr_path = '/DATA/ML_models/4_var/MLP'

# Number of samples to draw from the predictions for computing the firing rates
n_samples = 50
sim_params = {}
IDs = {}
firing_rates = {}

# Methods to plot
all_methods = ['catch22','power_spectrum_parameterization_1']

# Select the statistical analysis method ('cohen', 'lmer')
statistical_analysis = 'lmer'

# Random seed for numpy
np.random.seed(0)

all_IDs = {}
predictions_EI = {}
predictions_all = {}
ages = {}
for method in all_methods:
    # Load data
    try:
        data_EI = np.load(os.path.join(pred_results,method,'emp_data_reduced.pkl'), allow_pickle=True)
        data_all = np.load(os.path.join(pred_results,method,'emp_data_all.pkl'), allow_pickle=True)
        all_IDs[method] = np.array(data_all['ID'].tolist())
        predictions_EI[method] = np.array(data_EI['Predictions'].tolist())
        predictions_all[method] = np.array(data_all['Predictions'].tolist())
        ages[method] = np.array(data_EI['Group'].tolist())

        # Pick only ages >= 4
        all_IDs[method] = all_IDs[method][ages[method] >= 4]
        predictions_EI[method] = predictions_EI[method][ages[method] >= 4, :]
        predictions_all[method] = predictions_all[method][ages[method] >= 4, :]
        ages[method] = ages[method][ages[method] >= 4]

    except:
        all_IDs[method] = []
        predictions_EI[method] = []
        predictions_all[method] = []
        ages[method] = []

    firing_rates[method] = np.zeros((len(np.unique(ages[method])), n_samples))
    IDs[method] = np.zeros((len(np.unique(ages[method])), n_samples))

    # Parameter sampling for computing the firing rates
    if compute_firing_rate:
        sim_params[method] = np.zeros((7, len(np.unique(ages[method])), n_samples))
        for param in range(4):
            for i, age in enumerate(np.unique(ages[method])):
                idx = np.where(ages[method] == age)[0]
                data_IDs = all_IDs[method][idx]
                data_EI = predictions_EI[method][idx, param]
                data_EI = data_EI[~np.isnan(data_EI)]

                # Randomly sample some predictions within the first and third quartile
                q1, q3 = np.percentile(data_EI, [25, 75])

                # Check if the quartiles are not NaN
                if not np.isnan(q1) and not np.isnan(q3):
                    within_quartiles = np.where((data_EI >= q1) & (data_EI <= q3))[0]

                    # Check within_quartiles is not empty
                    if len(within_quartiles) > 0:
                        # Randomly sample n_samples from within_quartiles
                        idx_samples = within_quartiles[np.random.randint(0, len(within_quartiles), n_samples)]
                        IDs[method][i, :] = data_IDs[idx_samples]
                        # E/I
                        if param == 0:
                            for j in range(4):
                                data_all = predictions_all[method][idx, j]
                                data_all = data_all[~np.isnan(data_all)]
                                sim_params[method][j, i, :] = data_all[idx_samples]
                        # tau_syn_exc, tau_syn_inh, J_syn_ext
                        else:
                            sim_params[method][param+3, i, :] = data_EI[idx_samples]

        # Firing rates
        for i, age in enumerate(np.unique(ages[method])):
            for sample in range(n_samples):
                print(f'\nComputing firing rate for {method} at age {age} and sample {sample}')
                # Parameters of the model
                J_EE = sim_params[method][0, i, sample]
                J_IE = sim_params[method][1, i, sample]
                J_EI = sim_params[method][2, i, sample]
                J_II = sim_params[method][3, i, sample]
                tau_syn_E = sim_params[method][4, i, sample]
                tau_syn_I = sim_params[method][5, i, sample]
                J_ext = sim_params[method][6, i, sample]

                # Load LIF_params
                from network_params import LIF_params

                # Modify parameters
                LIF_params['J_YX'] = [[J_EE, J_IE], [J_EI, J_II]]
                LIF_params['tau_syn_YX'] = [[tau_syn_E, tau_syn_I],
                                            [tau_syn_E, tau_syn_I]]
                LIF_params['J_ext'] = J_ext

                # Create a Simulation object
                sim = ncpi.Simulation(param_folder='../../Hagen_model/simulation/params',
                                      python_folder='../../Hagen_model/simulation/python',
                                      output_folder='../../Hagen_model/simulation/output')

                # Save parameters to a pickle file
                with open(os.path.join('../../simulation/Hagen_model/simulation/output', 'network.pkl'), 'wb') as f:
                    pickle.dump(LIF_params, f)

                # Run the simulation
                sim.simulate('simulation.py', 'simulation_params.py')

                # Load spike times
                with open(os.path.join('../../simulation/Hagen_model/simulation/output', 'times.pkl'), 'rb') as f:
                    times = pickle.load(f)

                # Load tstop
                with open(os.path.join('../../simulation/Hagen_model/simulation/output', 'tstop.pkl'), 'rb') as f:
                    tstop = pickle.load(f)

                # Transient period
                from analysis_params import KernelParams
                transient = KernelParams.transient

                # Mean firing rate of excitatory cells
                times['E'] = times['E'][times['E'] >= transient]
                rate = ((times['E'].size / (tstop - transient)) * 1000) / LIF_params['N_X'][0]
                firing_rates[method][i, sample] = rate

        # Normalize firing rates to the maximum value
        if len(firing_rates[method]) > 0:
            if np.max(firing_rates[method]) > 0:
                firing_rates[method] /= np.max(firing_rates[method])

# Save firing rates to file
if compute_firing_rate:
    if not os.path.exists('data'):
        os.makedirs('data')
    with open('data/firing_rates_preds.pkl', 'wb') as f:
        pickle.dump(firing_rates, f)
    with open('data/IDs.pkl', 'wb') as f:
        pickle.dump(IDs, f)
else:
    try:
        with open(os.path.join(fr_path,'firing_rates_preds.pkl'), 'rb') as f:
            firing_rates = pickle.load(f)
        with open(os.path.join(fr_path,'IDs.pkl'), 'rb') as f:
            IDs = pickle.load(f)
    except FileNotFoundError:
        print('Firing rates not found.')
        pass

# Create a figure and set its properties
fig = plt.figure(figsize=(7.5, 5), dpi=300)
plt.rcParams.update({'font.size': 10, 'font.family': 'Arial'})
plt.rc('xtick', labelsize=8)
plt.rc('ytick', labelsize=8)

# Titles for the subplots
titles = [r'$E/I$', r'$\tau_{syn}^{exc}$ (ms)', r'$\tau_{syn}^{inh}$ (ms)',
          r'$J_{syn}^{ext}$ (nA)', r'$Norm. fr$']
# y-axis labels
y_labels = [r'$catch22$', r'$1/f$'+' '+r'$slope$']

# Define a colormap
cmap = plt.colormaps['viridis']

# Add rectangles to each row
for row in range(2):
    ax = fig.add_axes([0.01, 0.53 - row * 0.52, 0.98, 0.45 if row == 0 else 0.52])
    ax.add_patch(plt.Rectangle((0, 0), 1, 1, color='red' if row == 0 else 'blue', alpha=0.1))
    ax.set_xticks([])
    ax.set_yticks([])

# Plots
for row in range(2):
    for col in range(5):
        ax = fig.add_axes([0.08 + col * 0.19, 0.55 - row * 0.45, 0.14, 0.32])
        try:
            method = all_methods[row]
        except:
            method = all_methods[0]

        # Plot parameter predictions and firing rates as a function of age
        try:
            for i, age in enumerate(np.unique(ages[method])):
                idx = np.where(ages[method] == age)[0]
                if col < 4:
                    data_plot = predictions_EI[method][idx, col]
                else:
                    data_plot = firing_rates[method][i, :]

                # Remove NaNs
                data_plot = data_plot[~np.isnan(data_plot)]

                # # Boxplot
                # box = ax.boxplot(data_plot, positions=[age], showfliers=False,
                #                  widths=0.9, patch_artist=True, medianprops=dict(color='red', linewidth=0.8),
                #                  whiskerprops=dict(color='black', linewidth=0.5),
                #                  capprops=dict(color='black', linewidth=0.5),
                #                  boxprops=dict(linewidth=0.5))
                # for patch in box['boxes']:
                #     patch.set_facecolor(cmap(i / len(np.unique(ages[method]))))

                # Clip the data between the 5 % and 95 % quantiles
                q1, q3 = np.percentile(data_plot, [5, 95])
                clipped_data = data_plot[(data_plot >= q1) & (data_plot <= q3)]

                # Violin plot
                violin = ax.violinplot(clipped_data, positions=[age], widths=0.9, showextrema=False)

                for pc in violin['bodies']:
                    pc.set_facecolor(cmap(i / len(np.unique(ages[method]))))
                    pc.set_edgecolor('black')
                    pc.set_alpha(0.8)
                    pc.set_linewidth(0.2)

                # violin['cmedians'].set_linewidth(0.6)
                # violin['cmedians'].set_color('red')

                # Boxplot
                box = ax.boxplot(data_plot, positions=[age], showfliers=False,
                                 widths=0.5, patch_artist=True, medianprops=dict(color='red', linewidth=0.8),
                                 whiskerprops=dict(color='black', linewidth=0.5),
                                 capprops=dict(color='black', linewidth=0.5),
                                 boxprops=dict(linewidth=0.5, facecolor=(0, 0, 0, 0)))

                for patch in box['boxes']:
                    patch.set_linewidth(0.2)

                # # Debug: plot samples selected for the firing rate over the parameter predictions
                # if compute_firing_rate:
                #     if 0 < col < 4:
                #         ax.scatter([age]*n_samples, sim_params[method][col+3, i, :], color='black', s=0.5, zorder = 3)
                #     elif col == 0:
                #         ax.scatter([age]*n_samples, (sim_params[method][0, i, :]/sim_params[method][2, i, :]) /
                #                    (sim_params[method][1, i, :]/sim_params[method][3, i, :]),
                #                    color='black', s=2, zorder = 3)

            # stat. analysis
            if statistical_analysis == 'lmer':
                print('\n--- Linear mixed model analysis.')
            elif statistical_analysis == 'cohen':
                print('\n--- Cohen\'s d analysis.')

            data_EI = np.load(os.path.join(pred_results, method, 'emp_data_reduced.pkl'), allow_pickle=True)
            # Pick only ages >= 4
            data_EI = data_EI[data_EI['Group'] >= 4]

            # Create a DataFrame with the predictions
            if col < 4:
                data_EI['Y'] = predictions_EI[method][:, col]
            else:
                # Create a DataFrame with the firing rates
                data_fr = {'Group': np.repeat(np.unique(ages[method]), n_samples),
                           'ID': IDs[method].flatten(),
                           'Y': firing_rates[method].flatten(),
                           'Epoch': np.arange(firing_rates[method].size),
                           'Sensor': np.zeros(firing_rates[method].size)}
                data_EI = pd.DataFrame(data_fr)

            # Transform the 'Group' column to string type
            data_EI['Group'] = data_EI['Group'].astype(str)

            # Remove nan values from Y column
            data_EI = data_EI[~np.isnan(data_EI['Y'])]

            # Compute the statistical analysis
            Analysis = ncpi.Analysis(data_EI)
            if statistical_analysis == 'lmer':
                stat_result = Analysis.lmer(control_group='4', 
                                         data_col='Y',
                                         other_col=['ID', 'Group', 'Epoch', 'Sensor'],
                                         data_index=-1,
                                         models = {'mod00': 'Y ~ Group + (1 | ID)',
                                                             'mod01': 'Y ~ Group'},
                                         bic_models=["mod00", "mod01"],
                                         anova_tests = None,
                                         specs= '~Group')
            elif statistical_analysis == 'cohen':
                stat_result = Analysis.cohend(control_group='4', data_col='Y',data_index=-1)

            # Add p-values to the plot
            y_max = ax.get_ylim()[1]
            y_min = ax.get_ylim()[0]
            delta = (y_max - y_min) * 0.1

            # groups = ['5', '6', '7', '8', '9', '10', '11', '12']
            groups = ['8', '9', '10', '11', '12']
            for i, group in enumerate(groups):
                if statistical_analysis == 'lmer':
                    p_value = stat_result[f'{group}vs4']['p.value']
                    if p_value.empty:
                        continue
                elif statistical_analysis == 'cohen':
                    d_value = stat_result[f'{group}vs4']['d']
                    if d_value.empty:
                        continue

                # Significance levels
                if statistical_analysis == 'lmer':
                    if p_value.iloc[0] < 0.05 and p_value.iloc[0] >= 0.01:
                        pp = '*'
                    elif p_value.iloc[0] < 0.01 and p_value.iloc[0] >= 0.001:
                        pp = '**'
                    elif p_value.iloc[0] < 0.001 and p_value.iloc[0] >= 0.0001:
                        pp = '***'
                    elif p_value.iloc[0] < 0.0001:
                        pp = '****'
                    else:
                        pp = 'n.s.'

                    if pp != 'n.s.':
                        offset = -delta*0.2
                    else:
                        offset = delta*0.05

                elif statistical_analysis == 'cohen':
                    pp = f'{d_value[0]:.2f}'
                    offset = 0.

                ax.text((int(groups[i]) - 4)/2. + 4, y_max + delta*i + delta*0.1 + offset,
                        f'{pp}', ha='center', fontsize=8 if pp != 'n.s.' else 7)
                ax.plot([4, int(groups[0])+i], [y_max + delta*i, y_max + delta*i], color='black',
                        linewidth=0.5)

            # Change y-lim
            ax.set_ylim([y_min, y_max + delta*(len(groups))])

        except:
            pass

        # Titles
        ax.set_title(titles[col])

        # X-axis labels
        try:
            if row == 1:
                ax.set_xticks(np.unique(ages[method])[::2])
                ax.set_xticklabels([f'{str(i)}' for i in np.unique(ages[method])[::2]])
                ax.set_xlabel('Postnatal days')
            else:
                ax.set_xticks([])
                ax.set_xticklabels([])
        except:
            pass

        # # Y-axis labels
        # if col == 0:
        #     ax.set_ylabel(y_labels[row], color = 'red' if row == 0 else 'blue', alpha = 0.5)
        #     if row == 0:
        #         ax.yaxis.set_label_coords(-0.3, 0.5)
        #     else:
        #         ax.yaxis.set_label_coords(-0.3, 0.5)

# Plot letters
ax = fig.add_axes([0., 0., 1., 1.])
ax.axis('off')
ax.text(0.015, 0.945, 'A', fontsize=12, fontweight='bold')
ax.text(0.015, 0.495, 'B', fontsize=12, fontweight='bold')

# Titles
ax.text(0.5, 0.94, y_labels[0], color = 'red', alpha = 0.5, fontsize = 10, ha='center')
ax.text(0.5, 0.49, y_labels[1], color = 'blue', alpha = 0.5, fontsize = 10, ha='center')

# Save the figure
plt.savefig('LFP_predictions.png', bbox_inches='tight')
# plt.show()



import os
import numpy as np
from numpy.matlib import repmat # for compatibility with the original fE/I code
import pandas as pd
from scipy.signal import welch
import matplotlib.pyplot as plt
from ncpi import tools


class Features:
    """
    Class for computing features from electrophysiological data recordings.
    """

    def __init__(self, method='catch22', params=None):
        """
        Constructor method.

        Parameters
        ----------
        method: str
            Method to compute features. Default is 'catch22'.
        params: dict
            Dictionary containing the parameters for the feature computation.
        """

        # Assert that the method is a string
        if not isinstance(method, str):
            raise ValueError("The method must be a string.")

        # Check if the method is valid
        if method not in ['catch22', 'power_spectrum_parameterization', 'fEI', 'DFA', 'hctsa']:
            raise ValueError("Invalid method. Please use 'catch22', 'power_spectrum_parameterization', 'fEI', 'DFA', "
                             "or 'hctsa'.")

        # Import the required modules based on the method
        if method == 'catch22':
            if not tools.ensure_module("pycatch22"):
                raise ImportError("pycatch22 is required for computing catch22 features but is not installed.")
            self.pycatch22 = tools.dynamic_import("pycatch22")

        elif method == 'power_spectrum_parameterization':
            if not tools.ensure_module("fooof"):
                raise ImportError("fooof is required for computing power_spectrum_parameterization features but is "
                                  "not installed.")
            self.FOOOF = tools.dynamic_import("fooof", "FOOOF")

        elif method == 'fEI' or method == 'DFA':
            if not tools.ensure_module("PyAstronomy"):
                raise ImportError("PyAstronomy is required for computing fEI or DFA features but is not installed.")
            self.generalizedESD = tools.dynamic_import("PyAstronomy.pyasl", "generalizedESD")

        elif method == 'hctsa':
            if not tools.ensure_module("matlab"):
                raise ImportError("matlab is required for computing hctsa features but is not installed.")
            if not tools.ensure_module("matlabengine"):
                raise ImportError("matlabengine is required for computing hctsa features but is not installed.")
            if not tools.ensure_module("h5py"):
                raise ImportError("h5py is required for computing hctsa features but is not installed.")
            self.matlab = tools.dynamic_import("matlab")
            self.matlabengine = tools.dynamic_import("matlab","engine")
            self.h5py = tools.dynamic_import("h5py")

        # Check if params is a dictionary
        if not isinstance(params, dict) and params is not None:
            raise ValueError("params must be a dictionary.")

        self.method = method
        self.params = params

        # Check if pathos is installed. If not, use the default Python multiprocessing library
        if not tools.ensure_module("pathos"):
            self.pathos_inst = False
            self.multiprocessing = tools.dynamic_import("multiprocessing")
        else:
            self.pathos_inst = True
            self.pathos = tools.dynamic_import("pathos", "pools")

        # Check if tqdm is installed
        if not tools.ensure_module("tqdm"):
            self.tqdm_inst = False
        else:
            self.tqdm_inst = True
            self.tqdm = tools.dynamic_import("tqdm", "tqdm")

    def catch22(self,sample):
        """
        Compute the catch22 features from a time-series sample.

        Parameters
        ----------
        sample: np.array
            Sample data.

        Returns
        -------
        features: np.array
            Array with the catch22 feature values.
        """

        features = self.pycatch22.catch22_all(sample)

        return features['values']

    def power_spectrum_parameterization(self, sample, fs, fmin, fmax, fooof_setup, r_squared_th=0.9,
                                        freq_range=[30., 200.], nperseg=-1, compute_knee=False):
        """
        Power spectrum parameterization of a time-series sample using the FOOOF algorithm.

        Parameters
        ----------
        sample: np.array
            Times-series sample.
        fs: float
            Sampling frequency.
        fmin: float
            Minimum frequency for the power spectrum fit.
        fmax: float
            Maximum frequency for the power spectrum fit.
        fooof_setup: dict
            Dictionary containing the parameters for the FOOOF algorithm.
                - peak_threshold: float
                - min_peak_height: float
                - max_n_peaks: int
                - peak_width_limits: tuple
        r_squared_th: float
            Threshold for the r_squared value. Default is 0.9.
        freq_range: list
            Frequency range for the search of the peak parameters. Default is [30., 200.].
        nperseg: int
            Length of each segment for the power spectrum. Default is -1 (half a second).
        compute_knee: bool
            If True, the knee parameter is computed. Default is False.

        Returns
        -------
        features: np.array
            Array with the aperiodic components and peak parameters, plus the mean power:
            features[0:2] = aperiodic_params_fixed
            features[2:5] = peak_params_fixed
            features[5:8] = aperiodic_params_knee
            features[8:11] = peak_params_knee
            features[11] = mean power
        """

        debug = False
        features = np.full(12, np.nan)

        # Check that the length of the sample is at least 2 seconds
        if len(sample) >= 2 * fs:
            # Estimate power spectral density using Welch’s method
            if nperseg == -1:
                fxx, Pxx = welch(sample, fs, nperseg=int(0.5 * fs))
            else:
                fxx, Pxx = welch(sample, fs, nperseg=nperseg)

            if fmin >= fxx[0] and fmax <= fxx[-1]:
                f1 = np.where(fxx >= fmin)[0][0]
                f2 = np.where(fxx >= fmax)[0][0]
            else:
                print(
                    'Warning: fmin and fmax are out of the frequency range of the power spectrum. Adjusting fmin and fmax '
                    'to the minimum and maximum frequencies of the power spectrum.')
                f1 = fxx[0]
                f2 = fxx[-1]

            # Ensure the input data has no 0s
            if not np.any(Pxx == 0):
                # Fit the power spectrum using FOOOF for both aperiodic modes (fixed and knee)
                for ii, aperiodic_mode in enumerate(['fixed', 'knee'] if compute_knee else ['fixed']):
                    fm = self.FOOOF(peak_threshold=fooof_setup['peak_threshold'],
                               min_peak_height=fooof_setup['min_peak_height'],
                               max_n_peaks=fooof_setup['max_n_peaks'],
                               aperiodic_mode=aperiodic_mode,
                               peak_width_limits=fooof_setup['peak_width_limits'])
                    try:
                        fm.fit(fxx[f1:f2], Pxx[f1:f2])
                    except:
                        print('Error fitting the power spectrum.')
                        return np.full(12, np.nan)

                    # Discard fits with negative exponents
                    if fm.aperiodic_params_[-1] <= 0.:
                        fm.r_squared_ = 0.
                    # Discard nan r_squared
                    if np.isnan(fm.r_squared_):
                        fm.r_squared_ = 0.

                    # Print parameters and plot the fit
                    if debug:
                        print('fm.aperiodic_params_ = ', fm.aperiodic_params_)
                        print('fm.peak_params_ = ', fm.peak_params_)
                        print('fm.r_squared_ = ', fm.r_squared_)

                        fm.plot(plot_peaks='shade', peak_kwargs={'color': 'green'})

                        plt.title(f'aperiodic_params = {fm.aperiodic_params_}\n'
                                  f'peak_params = {fm.peak_params_}\n'
                                  f'r_squared = {fm.r_squared_}', fontsize=12)
                        plt.show()

                    # Collect the aperiodic and peak parameters
                    if fm.r_squared_ >= r_squared_th:
                        if ii == 0:
                            features[0:2] = fm.aperiodic_params_
                        else:
                            features[5:8] = fm.aperiodic_params_

                        if fm.peak_params_ is not None:
                            # Find peaks within the frequency range
                            pos_freq = np.where((fm.peak_params_[:, 0] >= freq_range[0]) &
                                                (fm.peak_params_[:, 0] <= freq_range[1]))[0]
                            # spectral shapes that have an oscillatory peak
                            if len(pos_freq) > 0:
                                peak_params = fm.peak_params_[pos_freq, :]

                                # Find the peak with the maximum power
                                pos_max = np.argmax(peak_params[:, 1])
                                peak_freq = peak_params[pos_max, 0]
                                peak_power = peak_params[pos_max, 1]
                                peak_BW = peak_params[pos_max, 2]

                                if ii == 0:
                                    features[2:5] = [peak_freq, peak_power, peak_BW]
                                else:
                                    features[8:11] = [peak_freq, peak_power, peak_BW]

                # Collect mean power
                features[11] = np.mean(Pxx[f1:f2])

        return features

    def _create_window_indices(self, length_signal, length_window, window_offset):

        """
        Function to create window indices for a signal.

        This code was downloaded from https://github.com/arthur-ervin/crosci/tree/main. This code is licensed under
        creative commons license CC-BY-NC https://creativecommons.org/licenses/by-nc/4.0/legalcode.txt.

        Parameters
        ----------
        length_signal: int
            Length of the signal.
        length_window: int
            Length of the window.
        window_offset: int
            Offset for the window.

        Returns
        -------
        all_window_index: np.array
            Array with the window indices.
        """

        window_starts = np.arange(0, length_signal - length_window, window_offset)
        num_windows = len(window_starts)

        one_window_index = np.arange(0, length_window)
        all_window_index = repmat(one_window_index, num_windows, 1).astype(int)

        all_window_index = all_window_index + repmat(np.transpose(window_starts[np.newaxis, :]), 1,
                                                     length_window).astype(int)

        return all_window_index

    def fEI(self, signal, sampling_frequency, window_size_sec, window_overlap, DFA_array, bad_idxes=[]):
        """
        Calculates fEI (on a set window size) for a signal

            Steps refer to description of fEI algorithm in Figure 2D of paper:
              Measurement of excitation inhibition ratio in autism spectrum disorder using critical brain dynamics
              Scientific Reports (2020)
              Hilgo Bruining*, Richard Hardstone*, Erika L. Juarez-Martinez*, Jan Sprengers*, Arthur-Ervin Avramiea,
              Sonja Simpraga, Simon J. Houtman, Simon-Shlomo Poil5, Eva Dallares, Satu Palva, Bob Oranje, J. Matias Palva,
              Huibert D. Mansvelder & Klaus Linkenkaer-Hansen
              (*Joint First Author)

        This code was downloaded from https://github.com/arthur-ervin/crosci/tree/main. This code is licensed under
        creative commons license CC-BY-NC https://creativecommons.org/licenses/by-nc/4.0/legalcode.txt.

        Originally created by Richard Hardstone (2020), rhardstone@gmail.com. Please note that commercial use of this
        algorithm is protected by Patent claim (PCT/NL2019/050167) “Method of determining brain activity”,
        with priority date 16 March 2018.


        Parameters
        ----------
        signal: array, shape(n_channels,n_times)
            amplitude envelope for all channels
        sampling_frequency: integer
            sampling frequency of the signal
        window_size_sec: float
            window size in seconds
        window_overlap: float
            fraction of overlap between windows (0-1)
        DFA_array: array, shape(n_channels)
            array of DFA values, with corresponding value for each channel, used for thresholding fEI
        bad_idxes: array, shape(n_channels)
            channels to ignore from computation are marked with 1, the rest with 0. can also be empty list,
            case in which all channels are computed

        Returns
        -------
        fEI_outliers_removed: array, shape(n_channels)
            fEI values, with outliers removed
        fEI_val: array, shape(n_channels)
            fEI values, with outliers included
        num_outliers: integer
            number of detected outliers
        wAmp: array, shape(n_channels, num_windows)
            windowed amplitude, computed across all channels/windows
        wDNF: array, shape(n_channels, num_windows)
            windowed detrended normalized fluctuation, computed across all channels/windows
        """

        window_size = int(window_size_sec * sampling_frequency)

        num_chans = np.shape(signal)[0]
        length_signal = np.shape(signal)[1]

        channels_to_ignore = [False] * num_chans

        for bad_idx in bad_idxes:
            channels_to_ignore[bad_idx] = True

        window_offset = int(np.floor(window_size * (1 - window_overlap)))
        all_window_index = _create_window_indices(length_signal, window_size, window_offset)
        num_windows = np.shape(all_window_index)[0]

        fEI_val = np.zeros((num_chans, 1))
        fEI_val[:] = np.NAN
        fEI_outliers_removed = np.zeros((num_chans, 1))
        fEI_outliers_removed[:] = np.NAN
        num_outliers = np.zeros((num_chans, 1))
        num_outliers[:] = np.NAN
        wAmp = np.zeros((num_chans, num_windows))
        wAmp[:] = np.NAN
        wDNF = np.zeros((num_chans, num_windows))
        wDNF[:] = np.NAN

        for ch_idx in range(num_chans):
            if channels_to_ignore[ch_idx]:
                continue

            original_amp = signal[ch_idx, :]

            if np.min(original_amp) == np.max(original_amp):
                print('Problem computing fEI for channel idx ' + str(ch_idx))
                continue

            signal_profile = np.cumsum(original_amp - np.mean(original_amp))
            w_original_amp = np.mean(original_amp[all_window_index], axis=1)

            x_amp = repmat(np.transpose(w_original_amp[np.newaxis, :]), 1, window_size)
            x_signal = signal_profile[all_window_index]
            x_signal = np.divide(x_signal, x_amp)

            # Calculate local trend, as the line of best fit within the time window
            _, fluc, _, _, _ = np.polyfit(np.arange(window_size), np.transpose(x_signal), deg=1, full=True)
            # Convert to root-mean squared error, from squared error
            w_detrendedNormalizedFluctuations = np.sqrt(fluc / window_size)

            fEI_val[ch_idx] = 1 - np.corrcoef(w_original_amp, w_detrendedNormalizedFluctuations)[0, 1]

            gesd_alpha = 0.05
            max_outliers_percentage = 0.025  # this is set to 0.025 per dimension (2-dim: wAmp and wDNF), so 0.05 is max
            # smallest value for max number of outliers is 2 for generalizedESD
            max_num_outliers = max(int(np.round(max_outliers_percentage * len(w_original_amp))), 2)
            outlier_indexes_wAmp = self.generalizedESD(w_original_amp, max_num_outliers, gesd_alpha)[1]
            outlier_indexes_wDNF = self.generalizedESD(w_detrendedNormalizedFluctuations, max_num_outliers, gesd_alpha)[1]
            outlier_union = outlier_indexes_wAmp + outlier_indexes_wDNF
            num_outliers[ch_idx, :] = len(outlier_union)
            not_outlier_both = np.setdiff1d(np.arange(len(w_original_amp)), np.array(outlier_union))
            fEI_outliers_removed[ch_idx] = 1 - np.corrcoef(w_original_amp[not_outlier_both], \
                                                           w_detrendedNormalizedFluctuations[not_outlier_both])[0, 1]

            wAmp[ch_idx, :] = w_original_amp
            wDNF[ch_idx, :] = w_detrendedNormalizedFluctuations

        fEI_val[DFA_array <= 0.6] = np.nan
        fEI_outliers_removed[DFA_array <= 0.6] = np.nan

        return (fEI_outliers_removed, fEI_val, num_outliers, wAmp, wDNF)

    def DFA(self, signal, sampling_frequency, fit_interval, compute_interval, overlap=True, bad_idxes=[]):
        """ Calculates DFA of a signal.

        This code was downloaded from https://github.com/arthur-ervin/crosci/tree/main. This code is licensed under
        creative commons license CC-BY-NC https://creativecommons.org/licenses/by-nc/4.0/legalcode.txt.

        Parameters
        ----------
        signal: array, shape(n_channels,n_times)
            amplitude envelope for all channels
        sampling_frequency: integer
            sampling frequency of the signal
        fit_interval: list, length 2
            interval (in seconds) over which the DFA exponent is fit. should be included in compute_interval
        compute_interval: list, length 2
            interval (in seconds) over which DFA is computed
        overlap: boolean
            if set to True, then windows are generated with an overlap of 50%
        bad_idxes: array, shape(n_channels)
            channels to ignore from computation are marked with 1, the rest with 0. can also be empty list,
            case in which all channels are computed

        Returns
        -------
        dfa_array, window_sizes, fluctuations, dfa_intercept
        dfa_array: array, shape(n_channels)
            DFA value for each channel
        window_sizes: array, shape(num_windows)
            window sizes over which the fluctuation function is computed
        fluctuations: array, shape(num_windows)
            fluctuation function value at each computed window size
        dfa_intercept: array, shape(n_channels)
            DFA intercept for each channel
        """

        num_chans, num_timepoints = np.shape(signal)

        channels_to_ignore = [False] * num_chans
        for bad_idx in bad_idxes:
            channels_to_ignore[bad_idx] = True

        length_signal = np.shape(signal)[1]

        assert fit_interval[0] >= compute_interval[0] and fit_interval[1] <= compute_interval[
            1], 'CalcInterval should be included in ComputeInterval'
        assert compute_interval[0] >= 0.1 and compute_interval[
            1] <= 1000, 'ComputeInterval should be between 0.1 and 1000 seconds'
        assert compute_interval[1] / sampling_frequency <= num_timepoints, \
            'ComputeInterval should not extend beyond the length of the signal'

        # compute DFA window sizes for the given CalcInterval
        window_sizes = np.floor(np.logspace(-1, 3, 81) * sampling_frequency).astype(
            int)  # %logspace from 0.1 seccond (10^-1) to 1000 (10^3) seconds

        # make sure there are no duplicates after rounding
        window_sizes = np.sort(np.unique(window_sizes))

        window_sizes = window_sizes[(window_sizes >= compute_interval[0] * sampling_frequency) & \
                                    (window_sizes <= compute_interval[1] * sampling_frequency)]

        dfa_array = np.zeros(num_chans)
        dfa_array[:] = np.NAN
        dfa_intercept = np.zeros(num_chans)
        dfa_intercept[:] = np.NAN
        fluctuations = np.zeros((num_chans, len(window_sizes)))
        fluctuations[:] = np.NAN

        if max(window_sizes) <= num_timepoints:
            for ch_idx in range(num_chans):
                if channels_to_ignore[ch_idx]:
                    continue

                signal_for_channel = signal[ch_idx, :]

                for i_window_size in range(len(window_sizes)):
                    if overlap == True:
                        window_overlap = 0.5
                    else:
                        window_overlap = 0

                    window_size = window_sizes[i_window_size]
                    window_offset = np.floor(window_size * (1 - window_overlap))
                    all_window_index = _create_window_indices(length_signal, window_sizes[i_window_size], window_offset)
                    # First we convert the time series into a series of fluctuations y(i) around the mean.
                    demeaned_signal = signal_for_channel - np.mean(signal_for_channel)
                    # Then we integrate the above fluctuation time series ('y').
                    signal_profile = np.cumsum(demeaned_signal)

                    x_signal = signal_profile[all_window_index]

                    # Calculate local trend, as the line of best fit within the time window -> fluc is the sum of
                    # squared residuals
                    _, fluc, _, _, _ = np.polyfit(np.arange(window_size), np.transpose(x_signal), deg=1, full=True)

                    # Peng's formula - Convert to root-mean squared error, from squared error
                    # det_fluc = np.sqrt(np.mean(fluc / window_size))
                    # Richard's formula
                    det_fluc = np.mean(np.sqrt(fluc / window_size))
                    fluctuations[ch_idx, i_window_size] = det_fluc

                # get the positions of the first and last window sizes used for fitting
                fit_interval_first_window = np.argwhere(window_sizes >= fit_interval[0] * sampling_frequency)[0][0]
                fit_interval_last_window = np.argwhere(window_sizes <= fit_interval[1] * sampling_frequency)[-1][0]

                # take the previous to the first window size if the difference between the lower end of fitting and
                # the previous window is no more than 1% of the lower end of fitting and if the difference between the lower
                # end of fitting and the previous window is less than the difference between the lower end of fitting and
                # the current first window
                if (np.abs(window_sizes[fit_interval_first_window - 1] / sampling_frequency - fit_interval[0]) <=
                        fit_interval[0] / 100):
                    if np.abs(window_sizes[fit_interval_first_window - 1] / sampling_frequency - fit_interval[0]) < \
                            np.abs(window_sizes[fit_interval_first_window] / sampling_frequency - fit_interval[0]):
                        fit_interval_first_window = fit_interval_first_window - 1

                x = np.log10(window_sizes[fit_interval_first_window:fit_interval_last_window + 1])
                y = np.log10(fluctuations[ch_idx, fit_interval_first_window:fit_interval_last_window + 1])
                model = np.polyfit(x, y, 1)
                dfa_intercept[ch_idx] = model[1]
                dfa_array[ch_idx] = model[0]

        return (dfa_array, window_sizes, fluctuations, dfa_intercept)

    def hctsa(self, samples, hctsa_folder, workers=32):
        """
        Compute hctsa features.

        Parameters
        ----------
        samples: ndarray/list of shape (n_samples, times-series length)
            A set of samples of time-series data.
        hctsa_folder: str
            Folder where hctsa is installed.
        workers: int
            Number of MATLAB workers of the parallel pool.

        Returns
        -------
        feats: list of shape (n_samples, n_features)
            hctsa features.

        Debugging
        ---------
        This function has been debugged by approximating results shown
        in https://github.com/benfulcher/hctsaTutorial_BonnEEG.
        """

        feats = []

        # start Matlab engine
        print("\n--> Starting Matlab engine ...")
        eng = self.matlabengine.start_matlab()

        try:
            # Remove hctsa file
            if os.path.isfile(os.path.join(hctsa_folder, 'HCTSA.mat')):
                os.remove(os.path.join(hctsa_folder, 'HCTSA.mat'))

            # Change to hctsa folder
            eng.cd(hctsa_folder)

            # Startup hctsa script
            print("\n--> hctsa startup ...")
            st = eng.startup(nargout=0)
            print(st)

            # Check if samples is a list and convert it to a numpy array
            if isinstance(samples, list):
                samples = np.array(samples)

            # Create the input variables in Matlab
            eng.eval(f'timeSeriesData = cell(1,{samples.shape[0]});', nargout=0)
            eng.eval(f'labels = cell(1,{samples.shape[0]});', nargout=0)
            eng.eval(f'keywords = cell(1,{samples.shape[0]});', nargout=0)

            # Transfer time-series data to Matlab workspace
            for s in range(samples.shape[0]):
                eng.workspace['aux'] = self.matlab.double(list(samples[s]))
                eng.eval('timeSeriesData{1,%s} = aux;' % (s + 1), nargout=0)

            # Fill in the other 2 Matlab structures with the index of the sample
            for s in range(samples.shape[0]):
                eng.eval('labels{1,%s} = \'%s\';' % (str(s + 1), str(s + 1)), nargout=0)
                eng.eval('keywords{1,%s} = \'%s\';' % (str(s + 1), str(s + 1)), nargout=0)

            # Save variables into a mat file
            eng.eval('save INP_ccpi_ts.mat timeSeriesData labels keywords;', nargout=0)

            # Load mat file
            eng.eval('load INP_ccpi_ts.mat;', nargout=0)

            # Initialize an hctsa calculation
            print("\n--> hctsa TS_Init ...")
            eng.TS_Init('INP_ccpi_ts.mat',
                        'hctsa',
                        self.matlab.logical([False, False, False]),
                        nargout=0)

            # Open a parallel pool of a specific size
            if workers > 1:
                eng.parpool(workers)

            # Compute features
            print("\n--> hctsa TS_Compute ...")
            # eng.TS_Compute(matlab.logical([True]),nargout = 0)
            eng.eval('TS_Compute(true);', nargout=0)

            # Load hctsa file
            f = self.h5py.File(os.path.join(hctsa_folder, 'HCTSA.mat'), 'r')
            TS_DataMat = np.array(f.get('TS_DataMat'))
            # TS_Quality = np.array(f.get('TS_Quality'))

            # Create the array of features to return
            print(f'\n--> Formatting {TS_DataMat.shape[0]} features...')
            for s in range(samples.shape[0]):
                feats.append(list(TS_DataMat[:, s]))

            # Stop Matlab engine
            print("\n--> Stopping Matlab engine ...")
            eng.quit()

        except Exception as e:
            print(f"An error occurred: {e}")
            # Stop Matlab engine
            print("\n--> Stopping Matlab engine ...")
            eng.quit()
            raise e

        return feats


    def compute_features(self, data, hctsa_folder = None):
        """
        Function to compute features from the data.

        Parameters
        ----------
        data: pd.DataFrame
            DataFrame containing the data. The time-series samples must be in the 'Data' column.
        hctsa_folder: str
            Folder where hctsa is installed (if method is 'hctsa').

        Returns
        -------
        data: pd.DataFrame
            DataFrame containing the data with the features appended.
        """

        def process_batch(batch_tuple):
            """
            Function to process a batch of samples.

            Parameters
            ----------
            batch: list
                List of samples.

            Returns
            -------
            features: list
                List of features.
            """

            batch_index, batch = batch_tuple
            # Normalize the batch
            if self.method != 'fEI':
                batch = [(sample - np.mean(sample)) / np.std(sample) for sample in batch]
            features = []

            # Compute features of each sample in the batch
            if self.method != 'hctsa':
                for sample in batch:
                    if self.method == 'catch22':
                        features.append(self.catch22(sample))
                    elif self.method == 'power_spectrum_parameterization':
                        features.append(self.power_spectrum_parameterization(sample,
                                                                        self.params['fs'],
                                                                        self.params['fmin'],
                                                                        self.params['fmax'],
                                                                        self.params['fooof_setup'],
                                                                        self.params['r_squared_th']))
                    elif self.method == 'DFA':
                        features.append(self.DFA(sample,
                                            self.params['fs'],
                                            self.params['fit_interval'],
                                            self.params['compute_interval'],
                                            self.params['overlap'],
                                            self.params['bad_idxes']))

                    elif self.method == 'fEI':
                        features.append(self.fEI(sample,
                                            self.params['fs'],
                                            self.params['window_size_sec'],
                                            self.params['window_overlap'],
                                            self.params['DFA_array'],
                                            self.params['bad_idxes']))

            # Compute hctsa for the whole batch to avoid starting the Matlab engine multiple times
            else:
                if self.method == 'hctsa':
                    features = self.hctsa(batch,hctsa_folder)

            return batch_index,features

        # Split the data into batches using the number of available CPUs
        num_cpus = os.cpu_count()
        if self.method == 'hctsa':
            factor = 0.5 # decrease this factor to avoid memory issues with MATLAB engine
        else:
            factor = 10 # more chunks than available CPUs (10 is a factor to update the progress bar more frequently)

        batch_size = len(data['Data']) // int(num_cpus*factor)
        if batch_size == 0:
            batch_size = 1
        batches = [(i, data['Data'][i:i + batch_size]) for i in range(0, len(data['Data']), batch_size)]

        # Compute the features in parallel using all available CPUs with pathos
        if self.pathos_inst:
            with self.pathos.ProcessPool(num_cpus) as pool:
                if self.tqdm_inst:
                    results = list(self.tqdm(pool.imap(process_batch, batches), total=len(batches),
                                        desc="Computing features"))
                else:
                    results = list(pool.imap(process_batch, batches))
        # If pathos is not installed, use the default Python multiprocessing library
        else:
            with self.multiprocessing.Pool(num_cpus) as pool:
                if self.tqdm_inst:
                    results = list(self.tqdm(pool.imap(process_batch, batches), total=len(batches),
                                        desc="Computing features"))
                else:
                    results = list(pool.imap(process_batch, batches))

        # Sort the features based on the original index
        results.sort(key=lambda x: x[0])
        features = [feature for _, batch_features in results for feature in batch_features]

        # Append the features to the DataFrame
        pd_feat = pd.DataFrame({'Features': features})
        data = pd.concat([data, pd_feat], axis=1)
        return data